# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
"""
Models
++++++

The ``Nodes`` models represent the actual calculation points of the
threedicore and derived information like the cells they lie in.

For examples how to query the ``Nodes`` model see :ref:`api-label`


"""

from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import numpy as np
import itertools

from threedigrid.orm.models import Model
from threedigrid.orm.fields import ArrayField
from threedigrid.orm.base.fields import BooleanArrayField
from threedigrid.orm.fields import PointArrayField
from threedigrid.orm.fields import BboxArrayField
from threedigrid.geo_utils import transform_xys
from threedigrid.numpy_utils import get_smallest_uint_dtype
from threedigrid.admin.nodes import exporters
from threedigrid.admin.nodes import subsets
from six.moves import zip


NODE_SUBSETS = {
    'node_type__eq': subsets.NODE_TYPE__EQ_SUBSETS,
    'node_type__in': subsets.NODE_TYPE__IN_SUBSETS,
}


class Nodes(Model):
    content_pk = ArrayField()
    seq_id = ArrayField()
    calculation_type = ArrayField()
    coordinates = PointArrayField()
    cell_coords = BboxArrayField()
    zoom_category = ArrayField()
    node_type = ArrayField()
    is_manhole = BooleanArrayField()
    sumax = ArrayField()
    drain_level = ArrayField()
    storage_area = ArrayField()
    dmax = ArrayField()
    initial_waterlevel = ArrayField()

    SUBSETS = NODE_SUBSETS

    @property
    def connectionnodes(self):
        return self._filter_as(ConnectionNodes, content_pk__ne=0)

    @property
    def manholes(self):
        return self._filter_as(Manholes, is_manhole=True)

    @property
    def added_calculationnodes(self):
        return self._filter_as(AddedCalculationNodes, content_pk=0)

    def __init__(self, *args, **kwargs):

        super(Nodes, self).__init__(*args, **kwargs)

        self._exporters = [
            exporters.NodesOgrExporter(self),
        ]

    @property
    def locations_2d(self):
        data = self.subset('2D_open_water').data
        # x0 = 0, y0 = 0
        # Translate
        # data['coordinates'][0] += x0
        # data['coordinates'][1] += y0

        # Return [[node_id, coordinate_x + x0, coordinate_y + y0]]
        return list(zip(
            data['id'].tolist(),
            data['coordinates'][0].tolist(),
            data['coordinates'][1].tolist()))


class AddedCalculationNodes(Nodes):
    pass


class EmbeddedNodes(Nodes):
    embedded_in = ArrayField()


class ConnectionNodes(Nodes):
    initial_waterlevel = ArrayField()


class Manholes(ConnectionNodes):
    bottom_level = ArrayField()
    display_name = ArrayField()
    surface_level = ArrayField()
    shape = ArrayField()
    width = ArrayField()
    manhole_indicator = ArrayField()


class Cells(Nodes):
    """
    Model that represents the threedicore calculation cells. ``Cells`` are
    a sub-class of ``Nodes`` because they share the same attributes.
    ``Cells``, however, also have a z_coordinate, the bottom level of
    the grid cell and cell_coords, the lower left and upper right coordinates
    of the cells extent.

    """

    z_coordinate = ArrayField()
    pixel_width = ArrayField()
    pixel_coords = BboxArrayField()

    def __init__(self, *args, **kwargs):

        super(Cells, self).__init__(*args, **kwargs)
        self._exporters = [
            exporters.CellsOgrExporter(self),
        ]

    @property
    def bounds(self):
        """
        :return: coordinates of the cell bounds (counter clockwise):
                 minx, miny, maxx, miny, maxx, maxy, minx, maxy
        """
        minx, miny, maxx, maxy = self.cell_coords
        return np.vstack(
            (minx, miny,
             maxx, miny,
             maxx, maxy,
             minx, maxy))

    def get_id_from_xy(self, x, y, xy_epsg_code=None, subset_name=None):
        """
        :param x: the x coordinate in xy_epsg_code
        :param y: the y coordinate in xy_epsg_code
        :param subset_name: filter on a subset of cells

        :return: numpy array with cell id's for x, y
        """

        if xy_epsg_code and xy_epsg_code != self.epsg_code:
            xy = transform_xys(
                np.array([x]), np.array([y]),
                xy_epsg_code, self.epsg_code).flatten()
        else:
            xy = [x, y]

        inst = self
        if subset_name:
            inst = self.subset(subset_name)
        id = inst.filter(cell_coords__contains_point=xy).id
        return id.tolist()

    def get_ids_from_pix_bbox(self, bbox, subset_name='2D_OPEN_WATER'):
        """
        :param x: the x coordinate in xy_epsg_code
        :param y: the y coordinate in xy_epsg_code
        :param subset_name: filter on a subset of cells

        :return: numpy array with cell id's for x, y
        """

        inst = self
        if subset_name:
            inst = self.subset(subset_name)
        id = inst.filter(pixel_coords__intersects_bbox=bbox).id
        return id.tolist()

    def get_nodgrid(self, pix_bbox, subset_name='2D_OPEN_WATER'):
        ids = np.array(
            self.get_ids_from_pix_bbox(pix_bbox, subset_name=subset_name),
            dtype=np.int32
        )
        nodgrid = create_nodgrid(
            self.pixel_coords[:],
            ids[:],
            int(pix_bbox[2] - pix_bbox[0]),
            int(pix_bbox[3] - pix_bbox[1]),
            int(pix_bbox[0]),
            int(pix_bbox[1])
        )
        return nodgrid[::-1, ::]

    def get_extent_pixels(self):
        """Determine the extent of the cells (in pixels)

        The returned bounding box is left-inclusive; cells cover the
        the half-open intervals [xmin, xmax) and [ymin, ymax).

        :return: tuple of xmin, ymin, xmax, ymax or None
        """
        coords = self.pixel_coords
        mask = ~np.any(coords == -9999, axis=0)
        if not np.any(mask):
            return
        coords = coords[:, mask]
        xmin = coords[0].min()
        ymin = coords[1].min()
        xmax = coords[2].max()
        ymax = coords[3].max()
        return xmin, ymin, xmax, ymax

    def iter_by_tile(self, width, height):
        """Iterate over groups of cells given a tile shape (in pixels).

        The tiles are always aligned to pixel (0, 0) so that a single grid cell
        never overlaps with multiple tiles. For the the same reason, the tile
        size should be an integer multiple of the maximum cell size.

        :param width: the width of the tile in pixels
        :param height: the height of the tile in pixels

        :yield: (xmin, ymin, xmax, ymax), cells
        """
        # determine the width of the largest cell
        cell_size = self.pixel_width.max()
        if width % cell_size != 0 or height % cell_size != 0:
            raise ValueError(
                "width and height should be a multiple of {}".format(cell_size)
            )

        # determine the total extent of the 2d nodes
        xmin, ymin, xmax, ymax = self.get_extent_pixels()

        # determine the lower left and upper right corners
        i1 = int(xmin // width)
        j1 = int(ymin // height)
        i2 = int(np.ceil(float(xmax) / width))
        j2 = int(np.ceil(float(ymax) / height))

        # yield the tiles
        for i, j in itertools.product(range(i1, i2), range(j1, j2)):
            x1, y1, x2, y2 = (
                i * width,
                j * height,
                (i + 1) * width,
                (j + 1) * height,
            )
            result = self.filter(
                pixel_coords__intersects_bbox=(x1 + 1, y1 + 1, x2 - 1, y2 - 1)
            )
            yield (x1, y1, x2, y2), result

    def __repr__(self):
        return "<orm cells instance of {}>".format(self.model_name)

    def __contenttype__(self):
        return 'cells'


class Grid(Model):
    """
    Implemented fields:

    - nodm: the horizintal index of the cell within its refinement level
    - nodn: the vertical index of the cell within its refinement level
    - nodk: the refinement level, 1 being the smallest cell

    They all have the same size as nodes and cells, which is why this
    model lives in the nodes/model module. In fact they are attributes
    of the cell coordinates
    """

    nodm = ArrayField()
    nodn = ArrayField()
    nodk = ArrayField()
    ip = ArrayField()
    jp = ArrayField()

    def __init__(self, *args, **kwargs):
        super(Grid, self).__init__(*args, **kwargs)
        self.class_kwargs["n2dtot"] = kwargs["n2dtot"]

    @property
    def n2dtot(self):
        return self.class_kwargs["n2dtot"]

    @property
    def dx(self):
        """Return size of the grid cell for each refinement level, in meters.
        """
        return self._datasource["dx"][:]

    @property
    def transform(self):
        """Return the transformation that maps pixel_coords to coordinates.

        The six returned values (a, b, c, d, e, f) define the (affine)
        transform between coordinates (x, y) and pixel indices (i, j)
        as follows::

        >>> x = a * i + b * j + c
        >>> y = d * i + e * j + f

        Note that for a 3Di grid, the vertical pixel size is positive, while
        for most raster files this is negative. This means that you should flip
        the vertical axis of the raster when using the pixel coordinates.
        """
        size = float(self._datasource["dxp"][()])
        origin_x = float(self._datasource["x0p"][()])
        origin_y = float(self._datasource["y0p"][()])
        return size, 0.0, origin_x, 0.0, size, origin_y

    def get_pixel_map(self, dem_pixelsize, dem_shape):
        """
        get the node grid to pixel map

        :param dem_pixelsize: pixelsize of the geo tiff
        :param dem_shape: shape of the numpy representation of the geo tiff

        :return: flipped array of the dem_shape that matches the geotiff
        """

        # Convert nod_grid to smallest uint type possible
        dtype = get_smallest_uint_dtype(maxval=self.n2dtot)
        grid_arr = np.zeros(dem_shape, dtype=dtype)

        # applies for for 2D nodes only
        _k = self.nodk[0:self.n2dtot + 1] - 1
        _m = self.nodm[0:self.n2dtot + 1] - 1
        _n = self.nodn[0:self.n2dtot + 1] - 1

        # the size in pixels of each grid cell
        _size = self.dx[_k] / dem_pixelsize
        # corresponding node index
        _ind = np.arange(0, self.n2dtot + 1, dtype='int')

        n_slice_start = np.array(_n * _size, dtype='int')
        n_slice_end = np.array((_n * _size) + _size, dtype='int')
        n_slice = list(zip(n_slice_start, n_slice_end))
        m_slice_start = np.array(_m * _size, dtype='int')
        m_slice_end = np.array((_m * _size) + _size, dtype='int')
        m_slice = list(zip(m_slice_start, m_slice_end))

        for ns, ms, idx in zip(n_slice, m_slice, _ind):
            axis_x = slice(*ns)
            axis_y = slice(*ms)
            grid_arr[axis_x, axis_y] = idx

        # flip upside-down to match geotiff
        return grid_arr[::-1, ::]


def create_nodgrid(pixel_coords, ids, width, height, offset_i, offset_j):
    grid_arr = np.zeros((height, width), dtype=np.int32)
    for id in ids:
        i0 = np.maximum(pixel_coords[0, id], offset_i) - offset_i
        i1 = np.minimum(pixel_coords[2, id], offset_i + width) - offset_i
        j0 = np.maximum(pixel_coords[1, id], offset_j) - offset_j
        j1 = np.minimum(pixel_coords[3, id], offset_j + height) - offset_j
        grid_arr[j0:j1, i0:i1] = id
    return grid_arr



# constants for exporter objects.  At this point there
# is a only one exporter type which depends on ogr. As ogr
# is an optional install the constants have been moved to a
# module of their own

from __future__ import absolute_import

try:
    from osgeo import ogr
except ImportError:
    ogr = None

if ogr is not None:
    OGR_FIELD_TYPE_MAP = {
        'int': ogr.OFTInteger,
        'str': ogr.OFTString,
        'real': ogr.OFTReal,
        'float': ogr.OFTReal,
    }

SHP_DRIVER_NAME = 'ESRI Shapefile'
GEO_PACKAGE_DRIVER_NAME = 'GPKG'
GEOJSON_DRIVER_NAME = 'GeoJSON'


CHANNELS_EXPORT_FIELDS = [
    'id',
    'content_pk',
    'code',
    'kcu',
    'calculation_type',
    'dist_calc_points',
    'connection_node_start_pk',
    'connection_node_end_pk',
    'discharge_coefficient'
]

PIPES_EXPORT_FIELDS = [
    'id',
    'content_pk',
    'display_name',
    'material',
    'kcu',
    'friction_value',
    'calculation_type',
    'sewerage_type',
    'cross_section_shape',
    'cross_section_width',
    'cross_section_height',
    'invert_level_start_point',
    'invert_level_end_point',
    'connection_node_start_pk',
    'connection_node_end_pk',
    'discharge_coefficient',
]

WEIRS_EXPORT_FIELDS = [
    'id',
    'content_pk',
    'display_name',
    'code',
    'kcu',
    'cross_section_shape',
    'cross_section_width',
    'cross_section_height',
    'crest_level',
    'crest_type',
    'sewerage',
    'discharge_coefficient_positive',
    'discharge_coefficient_negative',
    'friction_value',
    'friction_type',
    'connection_node_start_pk',
    'connection_node_end_pk',
]

CULVERT_EXPORT_FIELDS = [
    'id',
    'content_pk',
    'display_name',
    'code',
    'kcu',
    'cross_section_shape',
    'cross_section_width',
    'cross_section_height',
    'calculation_type',
    'friction_value',
    'friction_type',
    'dist_calc_points',
    'discharge_coefficient_positive',
    'discharge_coefficient_negative',
    'invert_level_start_point',
    'invert_level_end_point',
    'connection_node_start_pk',
    'connection_node_end_pk',
]

ORIFICES_EXPORT_FIELDS = [
    'id',
    'content_pk',
    'display_name',
    'kcu',
    'crest_level',
    'crest_type',
    'sewerage',
    'discharge_coefficient_negative',
    'discharge_coefficient_positive',
    'friction_type',
    'friction_value',
    'connection_node_start_pk',
    'connection_node_end_pk',
]

CONNECTION_NODES_EXPORT_FIELDS = [
    'content_pk',
    'initial_waterlevel',   # missing in prepare step
]

MANHOLE_EXPORT_FIELDS = [
    'id',
    'content_pk',
    'display_name',
    'calculation_type',
    'sumax',
    'bottom_level',
    'surface_level',
    'drain_level',
    'width',
    'shape',
    'storage_area',
    'initial_waterlevel',
]

PUMPS_EXPORT_FIELDS = [
    'id',
    'content_pk',
    'display_name',
    'start_level',
    'bottom_level',
    'lower_stop_level',
    'capacity',
    'type'
]

LEVEES_EXPORT_FIELDS = [
    'id',
    'crest_level',
    'max_breach_depth',
]

BREACHES_EXPORT_FIELDS = [
    "id",
    'content_pk',
    "levmat",
    "levbr",
    "kcu",
    "levl",
]

CELLS_EXPORT_FIELDS = [
    "id",
    'content_pk',
    "node_type",
    "sumax",
    "z_coordinate",
]

DEFAULT_EXPORT_FIELDS = {
    'Lines': 'ALL',
    'Pipes': PIPES_EXPORT_FIELDS,
    'Channels': CHANNELS_EXPORT_FIELDS,
    'Weirs': WEIRS_EXPORT_FIELDS,
    'Culverts': CULVERT_EXPORT_FIELDS,
    'Orifices': ORIFICES_EXPORT_FIELDS,
    'Nodes': 'ALL',
    'ConnectionNodes': CONNECTION_NODES_EXPORT_FIELDS,
    'Manholes': MANHOLE_EXPORT_FIELDS,
    'Cells': CELLS_EXPORT_FIELDS,
    'Grid': 'ALL',
    'Breaches': BREACHES_EXPORT_FIELDS,
    'Levees': LEVEES_EXPORT_FIELDS,
    'Pumps': PUMPS_EXPORT_FIELDS,
}



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
"""
Base fields
"""
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import

import numpy as np

from threedigrid.admin.constants import NO_DATA_VALUE
from threedigrid.admin.utils import PKMapper


class ArrayField:
    """
    Generic field that can be used to describe values
    to be retrieved from a Datasource.
    """

    @staticmethod
    def get_value(datasource, name, **kwargs):
        """
        Returns: the data from the datasource
                 or None if 'name' is not in the datasource

        Optional transforms can be done here.
        """
        if name in list(datasource.keys()):
            return datasource[name]

        return np.array([])

    def __repr__(self):
        return self.__class__.__name__


class BooleanArrayField(ArrayField):
    """
    Generic array field for boolean values.

    Because HDF5 does not support boolean datatype. No data fields are
    interpreted as False.
    """

    @staticmethod
    def get_value(datasource, name, **kwargs):
        """
        Returns: the data from the datasource
                 or None if 'name' is not in the datasource

        Optional transforms can be done here.
        """
        if name in list(datasource.keys()):
            data = datasource[name][:]
            data[data == NO_DATA_VALUE] = 0
            return data

        return np.array([])

    def __repr__(self):
        return self.__class__.__name__


class IndexArrayField(ArrayField):
    """
    Simple pointer
    """

    def __init__(self, to=None):
        self.to = to

    def __repr__(self):
        return self.__class__.__name__


class TimeSeriesArrayField(ArrayField):
    @staticmethod
    def get_value(datasource, name, **kwargs):
        timeseries_filter = kwargs.get("timeseries_filter", slice(None))
        timeseries_filter_to_use = timeseries_filter

        if isinstance(timeseries_filter, np.ndarray):
            if timeseries_filter.dtype == np.dtype(bool):
                # Convert to integer index array
                # to support h5py >= 3.1.0
                timeseries_filter_to_use = np.argwhere(
                    timeseries_filter).flatten()

        if (
            isinstance(timeseries_filter_to_use, np.ndarray)
            and len(datasource[name].shape) > 1
        ):
            v = datasource[name][timeseries_filter_to_use, :]
        else:
            v = datasource[name][timeseries_filter_to_use]
        if v.size > 0:
            return v
        return np.array([])

    def __repr__(self):
        return self.__class__.__name__


class TimeSeriesCompositeArrayField(TimeSeriesArrayField):
    """
    Field for composite arrays.

    Result (netCDF) files split their data into subsets, e.g. 1D and 2D.
    A composite field can be used to combine several data source fields
    into a single model field by specifying a composition dict. Example::

        LINE_COMPOSITE_FIELDS = {
            'au': ['Mesh1D_au', 'Mesh2D_au'],
            'u1': ['Mesh1D_u1', 'Mesh2D_u1'],
            'q': ['Mesh1D_q', 'Mesh2D_q']
        }

    """

    def __init__(self, meta=None):
        self._meta = meta

    def get_value(self, datasource, name, **kwargs):
        """
        :param datasource: a datasource object that can return data on
            __getitem__()
        :param name: the name of the section to read, e.g a HF5
            group or netCDF variable
        :param kwargs:
            timeseries_filter (optional): read only a slice of
                the time dimension
            model_name: name of the model the field belongs to.
                Is used for a reverse lookup of the composite fields
            lookup_index (optional): a numpy array that will be used
                to sort the values by this lookup index

        Returns: the data from the datasource
                 or None if 'name' is not in the datasource

        Optional transforms can be done here.
        """
        timeseries_filter = kwargs.get("timeseries_filter", slice(None))
        # timeseries_filter contains only [False, False,...]
        # (empty) slices pass the condition
        if isinstance(timeseries_filter, np.ndarray):
            if timeseries_filter.dtype == np.dtype(bool) and not np.any(
                timeseries_filter
            ):
                return np.array([])
        lookup_index = kwargs.get("lookup_index")
        values = []
        source_names = self._meta.composite_fields.get(name)

        timeseries_filter_to_use = timeseries_filter

        if isinstance(timeseries_filter, np.ndarray):
            if timeseries_filter.dtype == np.dtype(bool):
                # Convert to integer index array
                # to support h5py >= 3.1.0
                timeseries_filter_to_use = np.argwhere(
                    timeseries_filter).flatten()

        for source_name in source_names:
            if source_name not in list(datasource.keys()):
                continue

            if (
                isinstance(timeseries_filter_to_use, np.ndarray)
                and len(datasource[source_name].shape) > 1
            ):
                values.append(datasource[source_name][
                    timeseries_filter_to_use, :])
            else:
                values.append(datasource[source_name][
                    timeseries_filter_to_use])

        if not values:
            return np.array([])
        # combine the two source to a single source
        hs = np.hstack(values)
        axs = 1 if len(hs.shape) == 2 else 0
        hs = np.insert(hs, 0, 0, axis=axs)
        del values
        # sort the stacked array by lookup
        if lookup_index is not None:
            return hs[:, lookup_index]
        return hs

    def __repr__(self):
        return self.__class__.__name__


class TimeSeriesSubsetArrayField(TimeSeriesArrayField):
    """
    Field for subset arrays (for example only spanning the 2d section)
    """

    def __init__(self, source_name=None, size=None):
        self._source_name = source_name
        self._size = size

    def get_value(self, datasource, name, **kwargs):
        """
        :param datasource: a datasource object that can return data on
            __getitem__()
        :param name: the name of the section to read, e.g a HF5
            group or netCDF variable
        :param kwargs:
            timeseries_filter (optional): read only a slice of
                the time dimension
            subset_index: index array where to store the subset
                values
            lookup_index (optional): a numpy array that will be used
                to sort the values by this lookup index

        Returns: the data from the datasource or None if 'name' is not
            in the datasource

        Optional transforms can be done here.
        """
        timeseries_filter = kwargs.get("timeseries_filter", slice(None))
        subset_index = kwargs["subset_index"]
        # timeseries_filter contains only [False, False,...]
        # (empty) slices pass the condition
        if isinstance(timeseries_filter, np.ndarray):
            if timeseries_filter.dtype == np.dtype(bool) and not np.any(
                timeseries_filter
            ):
                return np.array([])

        timeseries_filter_to_use = timeseries_filter

        if isinstance(timeseries_filter, np.ndarray):
            if timeseries_filter.dtype == np.dtype(bool):
                # Convert to integer index array
                # to support h5py >= 3.1.0
                timeseries_filter_to_use = np.argwhere(
                    timeseries_filter).flatten()

        lookup_index = kwargs.get("lookup_index")
        if self._source_name not in list(datasource.keys()):
            return np.array([])
        source_data = datasource[self._source_name][
            timeseries_filter_to_use, :]
        shp = (source_data.shape[0], self._size)
        templ = np.zeros(shp, dtype=source_data.dtype)

        # Fix trash element mis-aligments
        if source_data.shape[1] == subset_index.shape[0] - 1:
            subset_index = subset_index[1:]

        templ[:, subset_index] = source_data

        # sort the stacked array by lookup
        if lookup_index is not None:
            return templ[:, lookup_index]
        return templ

    def __repr__(self):
        return self.__class__.__name__


def resolve_h5py_file_datasource(h5py_file, dotted_path):
    # Follow the 'path' with dots to get a datasource
    # in the h5py_file
    # For example:
    #    dotted_path = 'lines.id'
    res = h5py_file
    for key in dotted_path.split("."):
        res = res.get(key)
    return res[:]


class MappedSubsetArrayField(ArrayField):
    """
    Field for subset arrays (for example only spanning the 2d section)
    And also mapped -> using another ordering than in the hdf5 file.
    """

    def __init__(
        self,
        array_to_map=None,
        map_from_array=None,
        map_to_array=None,
        subset_filter=None,
        skip_if_datasource_present=False,
    ):
        """
        Use map_from_array and map_to_array to map array_to_map
        onto an new array with the name as specified for this field.

        :param array_to_map: the 'group.datasource' that is going to be mapped
        :map_from_array: the 'group.datasource' that is used to map from
        :map_to_array: the 'group.datasource' to map to
        :skip_if_datasource_present: skip the whole mapping process if the
           array is available on the default datasource

        For example:
            breach_id = MappedSubsetArrayField(
                array_to_map='breaches.id',
                map_from_array='breaches.levl', map_to_array='lines.id',
                subset_filter={
                    'lines.kcu':
                        subsets.KCU__IN_SUBSETS['POTENTIAL_BREACH'][0]})

            Maps breach.id array into lines for kcu=55 on lines using
            breach.levl -> lines.id to order breach.id

        Warning: the subset_filter needs to match the 'map_to_array' in length,
        so make sure they use the same h5py group.
        """
        self._array_to_map = array_to_map
        self._map_from_array = map_from_array
        self._map_to_array = map_to_array
        self._subset_filter = subset_filter
        self._skip_if_datasource_present = skip_if_datasource_present

    def get_value(self, datasource, name, **kwargs):
        # Skip the whole mapping if the dataset is
        # already present on the default datasource
        if self._skip_if_datasource_present and name in\
           list(datasource.keys()):
            data = datasource[name][:]
            data[data == NO_DATA_VALUE] = 0
            return data

        # Retrieve all specified arrays from
        # the h5py_file
        array_to_map = resolve_h5py_file_datasource(
            datasource._h5py_file, self._array_to_map
        )
        map_from_array = resolve_h5py_file_datasource(
            datasource._h5py_file, self._map_from_array
        )
        map_to_array = resolve_h5py_file_datasource(
            datasource._h5py_file, self._map_to_array
        )

        subset_key, subset_value = list(self._subset_filter.items())[0]

        subset_filter_array = resolve_h5py_file_datasource(
            datasource._h5py_file, subset_key
        )

        # Create a filter based on the specified subset_filter
        boolean_filter = subset_filter_array == subset_value

        new_shape = map_to_array.shape

        # Adjust for array's with 2 dimensions
        if len(array_to_map.shape) == 2:
            new_shape = (array_to_map.shape[0], new_shape[0])

        # Create an empty array that we are going to (partial) fill
        data = np.full(new_shape, 0, dtype=array_to_map.dtype)

        # Use the PKMapper to create a mapping from:
        # 'map_from_array' to 'map_to_array'
        # and apply that mapping
        # on the 'array_to_map'
        mapper = PKMapper(map_from_array, map_to_array[boolean_filter])

        output = mapper.apply_on(array_to_map)

        # Adjust filter dimensionality
        if len(array_to_map.shape) == 2:
            boolean_filter = tuple([slice(None), boolean_filter])

        # Fix trash element mis-aligments
        if output.shape[0] == data[boolean_filter].shape[0] + 1:
            output = output[1:]

        # Set the output on the correct subset on the data
        data[boolean_filter] = output
        return data



"""
Models
++++++

The ``Crosssections`` models represent the cross-sections as used in the
threedicore with shape, width and height information.


"""

from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import

from threedigrid.orm.models import Model
from threedigrid.orm.fields import ArrayField


class CrossSections(Model):
    code = ArrayField()
    shape = ArrayField()
    content_pk = ArrayField()
    width_1d = ArrayField()
    offset = ArrayField()
    count = ArrayField()
    tables = ArrayField()



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
"""
Models
++++++
In the threedicore lines are called flow lines and represent links between
calculation points. They can exist between 1d and 2d points, between surface
and groundwater nodes etc. The line type is described by the kcu attribute
For an overview of the kcu types see :ref:`kcu-label`.
"""

from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import numpy as np

from threedigrid.admin.lines import exporters
from threedigrid.orm.models import Model
from threedigrid.orm.fields import (
    ArrayField, LineArrayField, MultiLineArrayField)
from threedigrid.orm.base.fields import IndexArrayField, MappedSubsetArrayField
from threedigrid.admin.lines import subsets


LINE_SUBSETS = {
    'kcu__in': subsets.KCU__IN_SUBSETS
}


class Lines(Model):
    """
    fields originating from threedicore:

        - kcu (line type)
        - lik ()
        - line (index of node representing start/end point)

    added fields from spatialite database:

        - content_pk (primary key database)
        - content_type (type of object as stored in
          database, e.g v2_channel, v2_weir etc)

    custom fields
        - line_coords (coordinate pairs start/end point)
        - line_geometries (geometries from content_type)

    """
    kcu = ArrayField()
    lik = ArrayField()
    line = IndexArrayField(to='Nodes')
    dpumax = ArrayField()
    cross1 = IndexArrayField(to='CrossSections')
    cross2 = IndexArrayField(to='CrossSections')
    ds1d = ArrayField()
    ds1d_half = ArrayField()
    cross_weight = ArrayField()
    invert_level_start_point = ArrayField()
    invert_level_end_point = ArrayField()
    content_pk = ArrayField()
    content_type = ArrayField()
    zoom_category = ArrayField()
    cross_pix_coords = LineArrayField()
    line_coords = LineArrayField()
    line_geometries = MultiLineArrayField()
    SUBSETS = LINE_SUBSETS

    def __init__(self, *args,  **kwargs):
        super(Lines, self).__init__(*args, **kwargs)

        self._exporters = [
            exporters.LinesOgrExporter(self),
        ]

    @property
    def pipes(self):
        return self._filter_as(Pipes, content_type='v2_pipe')

    @property
    def weirs(self):
        return self._filter_as(Weirs, content_type='v2_weir')

    @property
    def orifices(self):
        return self._filter_as(Orifices, content_type='v2_orifice')

    @property
    def channels(self):
        return self._filter_as(Channels, content_type='v2_channel')

    @property
    def culverts(self):
        return self._filter_as(Culverts, content_type='v2_culvert')

    @property
    def breaches(self):
        return self._filter_as(Breaches, kcu=55)

    @property
    def line_nodes(self):
        return np.dstack(self.line)[0]


class Pipes(Lines):
    display_name = ArrayField()
    friction_type = ArrayField()
    friction_value = ArrayField()
    material = ArrayField()
    cross_section_height = ArrayField()
    cross_section_width = ArrayField()
    cross_section_shape = ArrayField()
    sewerage_type = ArrayField()
    calculation_type = ArrayField()
    connection_node_start_pk = ArrayField()
    connection_node_end_pk = ArrayField()
    discharge_coefficient = ArrayField()


class Channels(Lines):
    code = ArrayField()
    calculation_type = ArrayField()
    dist_calc_points = ArrayField()
    connection_node_start_pk = ArrayField()
    connection_node_end_pk = ArrayField()
    discharge_coefficient = ArrayField()


class Weirs(Lines):
    code = ArrayField()
    display_name = ArrayField()
    discharge_coefficient_negative = ArrayField()
    discharge_coefficient_positive = ArrayField()
    sewerage = ArrayField()
    friction_type = ArrayField()
    friction_value = ArrayField()
    crest_type = ArrayField()
    crest_level = ArrayField()
    connection_node_start_pk = ArrayField()
    connection_node_end_pk = ArrayField()
    cross_section_height = ArrayField()
    cross_section_width = ArrayField()
    cross_section_shape = ArrayField()

    @property
    def line_coord_angles(self):
        # Filter raw values with content_type and content_pk
        # to always get the raw (metric) values
        # and not the reprojected ones to calculate
        # the angles with.
        content_type = self._datasource['content_type'][:]
        line_content_pk = self._datasource['content_pk'][:]
        pk_mask = np.isin(line_content_pk, self.content_pk)
        raw_values = self._datasource['line_coords'][:][
            :, (content_type == b'v2_weir') & pk_mask]

        return self._get_field('line_coords').get_angles_in_degrees(
            raw_values)


class Culverts(Lines):
    display_name = ArrayField()
    code = ArrayField()
    calculation_type = ArrayField()
    dist_calc_points = ArrayField()
    cross_section_height = ArrayField()
    cross_section_width = ArrayField()
    cross_section_shape = ArrayField()
    friction_type = ArrayField()
    friction_value = ArrayField()
    discharge_coefficient_negative = ArrayField()
    discharge_coefficient_positive = ArrayField()
    connection_node_start_pk = ArrayField()
    connection_node_end_pk = ArrayField()


class Orifices(Lines):
    display_name = ArrayField()
    sewerage = ArrayField()
    friction_type = ArrayField()
    friction_value = ArrayField()
    crest_type = ArrayField()
    crest_level = ArrayField()
    discharge_coefficient_negative = ArrayField()
    discharge_coefficient_positive = ArrayField()
    connection_node_start_pk = ArrayField()
    connection_node_end_pk = ArrayField()


class Breaches(Lines):
    breach_id = MappedSubsetArrayField(
        array_to_map='breaches.id',
        map_from_array='breaches.levl', map_to_array='lines.id',
        subset_filter={
            'lines.kcu': subsets.KCU__IN_SUBSETS['POTENTIAL_BREACH'][0]})

    content_pk = MappedSubsetArrayField(
        array_to_map='breaches.content_pk',
        map_from_array='breaches.levl', map_to_array='lines.id',
        subset_filter={
            'lines.kcu': subsets.KCU__IN_SUBSETS['POTENTIAL_BREACH'][0]})

    levbr = MappedSubsetArrayField(
        array_to_map='breaches.levbr',
        map_from_array='breaches.levl', map_to_array='lines.id',
        subset_filter={
            'lines.kcu': subsets.KCU__IN_SUBSETS['POTENTIAL_BREACH'][0]})

    levmat = MappedSubsetArrayField(
        array_to_map='breaches.levmat',
        map_from_array='breaches.levl', map_to_array='lines.id',
        subset_filter={
            'lines.kcu': subsets.KCU__IN_SUBSETS['POTENTIAL_BREACH'][0]})

    coordinates = MappedSubsetArrayField(
        array_to_map='breaches.coordinates',
        map_from_array='breaches.levl', map_to_array='lines.id',
        subset_filter={
            'lines.kcu': subsets.KCU__IN_SUBSETS['POTENTIAL_BREACH'][0]})



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import numpy as np
from .utils import PKMapper
import six

DEFAULT_NULL_VALUE = -9999

FIELD_NULL_VALUES = {
    'display_name':  b'',
    'code': b'',
}


def db_objects_to_numpy_array_dict(db_objects, field_names):
    numpy_array_dict = {}

    for field_name in field_names:
        numpy_array_dict[field_name] = [DEFAULT_NULL_VALUE] * len(db_objects)

    for index, db_object in enumerate(db_objects):
        for field_name in field_names:
            if '__' in field_name:
                # Allow to get attrs of child objects
                # by using field__attr
                splitted = field_name.split('__')
                db_object_attr = getattr(db_object, splitted[0])
                value = getattr(db_object_attr, splitted[1])
            else:
                value = getattr(db_object, field_name)

            if field_name == 'the_geom':
                value = np.fromstring(str(value.wkb), dtype='uint8')

            # Convert unicode to str
            if isinstance(value, six.text_type):
                value = value.encode('utf8')
            if value is not None:
                numpy_array_dict[field_name][index] = value

    for key in numpy_array_dict:
        numpy_array_dict[key] = np.array(numpy_array_dict[key])

    return numpy_array_dict


def add_or_update_datasets(h5py_group, numpy_array_dict, field_names,
                           pk, content_pk, ignore_mask=None,
                           field_name_override=None):

    # map pk onto content_pk
    pk_mapper = PKMapper(pk, content_pk, ignore_mask)
    for field_name in field_names:
        if field_name == 'pk':
            # Never store pk
            continue

        np_array = numpy_array_dict[field_name]

        null_value = FIELD_NULL_VALUES.get(
            field_name, DEFAULT_NULL_VALUE)

        data = pk_mapper.apply_on(
            np_array, null_value)

        if field_name_override:
            # By default use field_name if there is no override
            dataset_name = field_name_override.get(field_name, field_name)
        else:
            dataset_name = field_name

        if dataset_name not in list(h5py_group.keys()):
            dt = None
            if data.dtype.type is np.string_:
                if dataset_name == 'code':
                    dt = 'S32'
                elif dataset_name == 'display_name':
                    dt = 'S64'
                elif dataset_name == 'shape':
                    dt = 'S4'

            h5py_group.create_dataset(
                dataset_name, data=data, dtype=dt)
        else:
            values = h5py_group[dataset_name][:]
            mask = data != null_value
            values[mask] = data[mask]
            h5py_group[dataset_name][:] = values

    del data



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import

import json
import logging
import os

import numpy as np
import six

from threedigrid.admin.h5py_datasource import H5pyGroup
from threedigrid.admin.idmapper import IdMapper
from threedigrid.admin import constants
from threedigrid.admin.gridadmin import GridH5Admin

from threedigrid.admin.breaches.prepare import PrepareBreaches
from threedigrid.admin.lines.prepare import (
    PrepareLines, PrepareChannels, PreparePipes, PrepareWeirs,
    PrepareOrifices, PrepareCulverts)

from threedigrid.admin.levees.prepare import PrepareLevees
from threedigrid.admin.nodes.prepare import (
    PrepareNodes, PrepareManholes, PrepareConnectionNodes, PrepareCells)
from threedigrid.admin.pumps.prepare import PreparePumps
from threedigrid.admin.levees.models import Levees
from threedigrid.orm.base.encoder import NumpyEncoder
from threedigrid.orm.constants import EXPORT_METHOD_TO_EXTENSION_MAP

logger = logging.getLogger(__name__)


def is_prepared(h5py_file, group_name, attr_name):
    if group_name not in h5py_file:
        return False
    return h5py_file[group_name].attrs.get(attr_name, 0) == 1


def skip_prepare(h5py_file, group_name, attr_name, overwrite):
    if overwrite:
        return False
    return is_prepared(h5py_file, group_name, attr_name)


def prepare_lines_onedee(
        h5py_file, threedi_datasource, klass, attr_name, overwrite=False):

    if skip_prepare(h5py_file, 'lines', attr_name, overwrite):
        return

    klass.prepare_datasource(h5py_file, threedi_datasource)

    h5py_file['lines'].attrs[attr_name] = 1


class GridAdminH5Prepare(object):
    @staticmethod
    def prepare(h5py_file, threedi_datasource, extra_attrs=None):
        # Step 1: Set optional extra attrs like epsg code,
        # model_name, revision_number, revision_hash
        # and possible other meta data..
        if extra_attrs:
            for key, value in six.iteritems(extra_attrs):
                if isinstance(value, six.string_types):
                    value = np.string_(value)
                h5py_file.attrs[key] = value

        # Step 3: prepare the ID mapper
        IdMapper.prepare_mapper(
            h5py_file, threedi_datasource)

        # Step 4: Prepare nodes/lines
        GridAdminH5Prepare.prepare_nodes(h5py_file, threedi_datasource)
        GridAdminH5Prepare.prepare_lines(h5py_file, threedi_datasource)

        # Step 5: Prepare onedee lines/nodes/pumps
        GridAdminH5Prepare.prepare_onedee_lines(h5py_file, threedi_datasource)
        GridAdminH5Prepare.prepare_onedee_nodes(h5py_file, threedi_datasource)
        GridAdminH5Prepare.prepare_pumps(h5py_file, threedi_datasource)

        # Step 6: prepare levees
        GridAdminH5Prepare.prepare_levees(h5py_file, threedi_datasource)

        # Step 7: prepare breaches
        GridAdminH5Prepare.prepare_breaches(h5py_file, threedi_datasource)

    @staticmethod
    def prepare_levees(h5py_file, threedi_datasource, overwrite=False):
        if not threedi_datasource.levees:
            return

        if skip_prepare(h5py_file, 'levees', 'prepared', overwrite):
            return

        PrepareLevees.prepare_datasource(h5py_file, threedi_datasource)

        h5py_file['levees'].attrs['prepared'] = 1

    @staticmethod
    def prepare_onedee_nodes(h5py_file, threedi_datasource, overwrite=False):
        has_1d = h5py_file.attrs.get('has_1d', 0) == 1

        if not has_1d:
            return

        if not is_prepared(h5py_file, 'nodes', 'prepared'):
            # Line data is missing, prepare line data first.
            GridAdminH5Prepare.prepare_nodes(h5py_file, threedi_datasource)

        if skip_prepare(h5py_file, 'nodes', 'connectionnodes_prepared',
                        overwrite):
            return

        PrepareConnectionNodes.prepare_datasource(
                h5py_file, threedi_datasource)

        h5py_file['nodes'].attrs['connectionnodes_prepared'] = 1

        if skip_prepare(h5py_file, 'nodes', 'manholes_prepared', overwrite):
            return

        PrepareManholes.prepare_datasource(h5py_file, threedi_datasource)

        h5py_file['nodes'].attrs['manholes_prepared'] = 1

    @staticmethod
    def prepare_nodes(h5py_file, threedi_datasource, overwrite=False):
        if skip_prepare(h5py_file, 'nodes', 'prepared', overwrite):
            return

        has_1d = h5py_file.attrs.get('has_1d', 0) == 1

        if has_1d:
            mapping1d = H5pyGroup(h5py_file, 'mapping1d')
        else:
            mapping1d = None

        id_mapper = IdMapper(H5pyGroup(h5py_file, 'mappings')['id_map'])

        # Prepare nodes
        node_group = H5pyGroup(h5py_file, 'nodes')
        PrepareNodes.prepare_datasource(
            node_group, mapping1d, id_mapper, has_1d=has_1d)

        PrepareCells.prepare_datasource(h5py_file, threedi_datasource)

        h5py_file['nodes'].attrs['prepared'] = 1

    @staticmethod
    def prepare_lines(h5py_file, threedi_datasource, overwrite=False):
        # Prepare lines
        if skip_prepare(h5py_file, 'lines', 'lines_prepared', overwrite):
            return

        has_1d = h5py_file.attrs.get('has_1d', 0) == 1

        line_group = H5pyGroup(h5py_file, 'lines')
        node_group = H5pyGroup(h5py_file, 'nodes')

        id_mapper = IdMapper(H5pyGroup(h5py_file, 'mappings')['id_map'])

        PrepareLines.prepare_datasource(
            line_group,
            id_mapper, threedi_datasource,
            [node_group['x_coordinate'], node_group['y_coordinate']],
            has_1d=has_1d
        )

        h5py_file['lines'].attrs['lines_prepared'] = 1

    @staticmethod
    def prepare_breaches(h5py_file, threedi_datasource, overwrite=True):
        if skip_prepare(h5py_file, 'breaches', 'prepared', overwrite):
            return

        if not is_prepared(h5py_file, 'lines', 'line_prepared'):
            # Line data is missing, prepare line data first.
            GridAdminH5Prepare.prepare_lines(h5py_file, threedi_datasource)

        if not is_prepared(h5py_file, 'levees', 'prepared'):
            # Levee data is missing, prepare levee data first.
            GridAdminH5Prepare.prepare_levees(h5py_file, threedi_datasource)

        has_breaches = h5py_file.attrs.get('has_breaches', 0) == 1
        levees = None
        if 'levees' in h5py_file:
            levees = Levees(H5pyGroup(h5py_file, 'levees'))

        id_mapper = IdMapper(H5pyGroup(h5py_file, 'mappings')['id_map'])
        line_group = H5pyGroup(h5py_file, 'lines')

        # Prepare breaches
        if has_breaches:
            PrepareBreaches.prepare_datasource(
                H5pyGroup(h5py_file, 'breaches', required=True),
                line_group['kcu'], id_mapper, levees,
                line_group['line_coords']
            )

            h5py_file['breaches'].attrs['prepared'] = 1

    @staticmethod
    def prepare_onedee_lines(h5py_file, threedi_datasource, overwrite=False):
        has_1d = h5py_file.attrs.get('has_1d', 0) == 1

        if not has_1d:
            return

        if not is_prepared(h5py_file, 'lines', 'lines_prepared'):
            # Line data is missing, prepare line data first.
            GridAdminH5Prepare.prepare_lines(h5py_file, threedi_datasource)

        # Load data from threedi_datasource into the h5py_file on
        # the lines group.
        prepare_lines_onedee(
            h5py_file, threedi_datasource, PrepareChannels,
            'channels_prepared', overwrite)

        prepare_lines_onedee(
            h5py_file, threedi_datasource, PreparePipes,
            'pipes_prepared', overwrite)

        prepare_lines_onedee(
            h5py_file, threedi_datasource, PrepareWeirs,
            'weirs_prepared', overwrite)

        prepare_lines_onedee(
            h5py_file, threedi_datasource, PrepareOrifices,
            'orifices_prepared', overwrite)

        prepare_lines_onedee(
            h5py_file, threedi_datasource, PrepareCulverts,
            'culverts_prepared', overwrite)

    @staticmethod
    def prepare_pumps(h5py_file, threedi_datasource, overwrite=False):
        has_1d = h5py_file.attrs.get('has_1d', 0) == 1
        has_pumpstations = h5py_file.attrs.get('has_pumpstations', 0) == 1

        if not has_1d or not has_pumpstations:
            return

        if skip_prepare(h5py_file, 'pumps', 'prepared', overwrite):
            return

        # if not h5py_file.attrs.get('has_pumpstations'):
        #    logger.info('[*] Datasource does not have pumps, skipping...')
        #    return

        if not is_prepared(h5py_file, 'nodes', 'prepared'):
            # Node data is missing, prepare node data first.
            GridAdminH5Prepare.prepare_nodes(h5py_file, threedi_datasource)

        PreparePumps.prepare_datasource(
            h5py_file, threedi_datasource, H5pyGroup(h5py_file, 'pumps'))

        h5py_file['pumps'].attrs['prepared'] = 1


class GridAdminH5Export(object):
    """
    Exporter that saves shapefiles/geopackage/geojson to destination folder
    """

    def __init__(
            self,
            gridadmin_file,
            export_method='to_shape',
            target_epsg_code=None,
            destination=None,
            indent=None
    ):
        """

        :param gridadmin_file: hdf5 file (full path)
        :param export_method: (str) 'to_shape', 'to_gpkg' or 'to_geojson'
        :param target_epsg_code: (str) epsg code
        :param destination: path to folder the export folder, defaults to
            the same folder as the gridadmin_file
        """
        self.ga = GridH5Admin(gridadmin_file)
        self._export_method = export_method
        self._extension = EXPORT_METHOD_TO_EXTENSION_MAP[export_method]
        if not target_epsg_code:
            self._epsg = self.ga.epsg_code
        else:
            self._epsg = target_epsg_code
        if not destination:
            self._dest = os.path.split(gridadmin_file)[0]
        else:
            self._dest = destination
        self._indent = indent

    def export_all(self):
        """convenience function to run all exports"""
        self.export_nodes()
        self.export_lines()
        self.export_grid()
        self.export_levees()
        self.export_2d_groundwater_lines()
        self.export_2d_openwater_lines()
        self.export_2d_vertical_infiltration_lines()

    def export_frontend(self):
        self.export_breaches()
        self.export_channels()
        self.export_pipes()
        self.export_weirs()
        self.export_culverts()
        self.export_orifices()
        self.export_manholes()
        self.export_nodes()
        self.export_pumps()
        self.export_levees()
        self.export_grid()
        self._combine(
            output_name="all",
            file_names={
                'breaches', 'channels', 'pipes', 'weirs', 'culverts',
                'orifices', 'manholes', constants.NODES, 'pumps', 'levees'
            }
        )
        self._combine(
            output_name="cells",
            file_names={constants.OPEN_WATER, constants.GROUNDWATER}
        )

    def _combine(self, output_name, file_names):
        """Combine the `file_names` into a new file called `output_name`.

        The files are simply appended. Only works for geojson.

        :param output_name: name of the new file
        :param file_names: file names to combine
        :return:
        """
        if self._extension not in ('.json', '.geojson'):
            logger.info("Can only combine exports with geojson")
            return

        dest = os.path.join(self._dest, output_name + self._extension)
        with open(dest, 'w') as output_file:
            features = []
            for file_name in file_names:
                intermediate_result = os.path.join(
                    self._dest, file_name + self._extension
                )
                if not os.path.exists(intermediate_result):
                    continue

                with open(intermediate_result, 'r') as ir:
                    geojson = json.load(ir)
                    im_features = geojson['features']
                    features += im_features

            json.dump(
                {"type": "FeatureCollection", "features": features},
                output_file,
                indent=self._indent,
                cls=NumpyEncoder,
            )

    def export_1d_all(self):
        self.export_nodes()
        self.export_lines()
        self.export_levees()

    def export_nodes(self):
        """
        writes shapefile of all 1D nodes
        """
        if not self.ga.has_1d:
            logger.info(
                "[*] Model {} does not have 1D, "
                "skipping export nodes...".format(self.ga.model_name)
            )
            return
        dest = os.path.join(self._dest, constants.NODES + self._extension)
        getattr(
            self.ga.nodes.subset(
                constants.SUBSET_1D_ALL
            ).reproject_to(self._epsg),
            self._export_method
        )(dest, indent=self._indent)

    def export_lines(self):
        """
        writes shapefile of all 1D lines
        """
        if not self.ga.has_1d:
            logger.info(
                "[*] Model {} does not have 1D, "
                "skipping export lines...".format(self.ga.model_name)
            )
            return
        dest = os.path.join(self._dest, constants.LINES + self._extension)
        getattr(
            self.ga.lines.subset(
                constants.SUBSET_1D_ALL
            ).reproject_to(self._epsg),
            self._export_method
        )(dest, indent=self._indent)

    def export_breaches(self):
        if self.ga.breaches.id.size == 0:
            logger.info(
                "[*] Model {} does not have 1D, "
                "skipping export breaches...".format(self.ga.model_name)
            )
            return
        dest = os.path.join(self._dest, 'breaches' + self._extension)
        getattr(
            self.ga.breaches.reproject_to(self._epsg),
            self._export_method
        )(dest, indent=self._indent)

    def export_channels(self):
        if self.ga.lines.channels.id.size == 0:
            logger.info(
                "[*] Model {} does not have 1D, "
                "skipping export channels...".format(self.ga.model_name)
            )
            return
        dest = os.path.join(self._dest, 'channels' + self._extension)
        getattr(
            self.ga.lines.channels.reproject_to(self._epsg),
            self._export_method
        )(dest, indent=self._indent)

    def export_pipes(self):
        if self.ga.lines.pipes.id.size == 0:
            logger.info(
                "[*] Model {} does not have 1D, "
                "skipping export pipes...".format(self.ga.model_name)
            )
            return
        dest = os.path.join(self._dest, 'pipes' + self._extension)
        getattr(
            self.ga.lines.pipes.reproject_to(self._epsg),
            self._export_method
        )(dest, indent=self._indent)

    def export_weirs(self):
        if self.ga.lines.weirs.id.size == 0:
            logger.info(
                "[*] Model {} does not have 1D, "
                "skipping export weirs...".format(self.ga.model_name)
            )
            return
        dest = os.path.join(self._dest, 'weirs' + self._extension)
        getattr(
            self.ga.lines.weirs.reproject_to(self._epsg),
            self._export_method
        )(dest, indent=self._indent)

    def export_culverts(self):
        if self.ga.lines.culverts.id.size == 0:
            logger.info(
                "[*] Model {} does not have 1D, "
                "skipping export culverts...".format(self.ga.model_name)
            )
            return
        dest = os.path.join(self._dest, 'culverts' + self._extension)
        getattr(
            self.ga.lines.culverts.reproject_to(self._epsg),
            self._export_method
        )(dest, indent=self._indent)

    def export_orifices(self):
        if self.ga.lines.orifices.id.size == 0:
            logger.info(
                "[*] Model {} does not have 1D, "
                "skipping export orifices...".format(self.ga.model_name)
            )
            return
        dest = os.path.join(self._dest, 'orifices' + self._extension)
        getattr(
            self.ga.lines.orifices.reproject_to(self._epsg),
            self._export_method
        )(dest, indent=self._indent)

    def export_manholes(self):
        if self.ga.nodes.manholes.id.size == 0:
            logger.info(
                "[*] Model {} does not have 1D, "
                "skipping export manholes...".format(self.ga.model_name)
            )
            return
        dest = os.path.join(self._dest, 'manholes' + self._extension)
        getattr(
            self.ga.nodes.manholes.reproject_to(self._epsg),
            self._export_method
        )(dest, indent=self._indent)

    def export_pumps(self):
        if self.ga.pumps.id.size == 0:
            logger.info(
                "[*] Model {} does not have 1D, "
                "skipping export pumps...".format(self.ga.model_name)
            )
            return
        dest = os.path.join(self._dest, 'pumps' + self._extension)
        getattr(
            self.ga.pumps.reproject_to(self._epsg),
            self._export_method
        )(dest, indent=self._indent)

    def export_levees(self):
        """
        writes shapefile of levees
        """
        if not self.ga.has_levees:
            logger.info(
                "[*] Model {} does not have levees, "
                "skipping export levees...".format(self.ga.model_name)
            )
            return
        dest = os.path.join(self._dest, constants.LEVEES + self._extension)
        getattr(
            self.ga.levees.reproject_to(self._epsg),
            self._export_method
        )(dest, indent=self._indent)

    def export_grid(self):
        """
        writes shapefile of all 2D groundwater and open water
        cells (if present)
        """
        self._export_2d_grid()
        self._export_groundwater_grid()

    def export_2d_groundwater_lines(self):
        if not (self.ga.has_groundwater and self.ga.has_2d):
            logger.info(
                "[*] Model {} does not have 2d groundwater lines, "
                "skipping export 2d groundwater lines...".format(
                    self.ga.model_name)
            )
            return
        dest = os.path.join(
            self._dest, constants.GROUNDWATER_LINES + self._extension
        )
        getattr(
            self.ga.lines.subset(
                constants.SUBSET_2D_GROUNDWATER).reproject_to(self._epsg),
            self._export_method
        )(dest, indent=self._indent)

    def export_2d_openwater_lines(self):
        if not self.ga.has_2d:
            logger.info(
                "[*] Model {} does not have 2D, "
                "skipping export 2d open water lines...".format(
                    self.ga.model_name)
            )
            return
        dest = os.path.join(
            self._dest, constants.OPEN_WATER_LINES + self._extension
        )
        getattr(
            self.ga.lines.subset(
                constants.SUBSET_2D_OPEN_WATER).reproject_to(self._epsg),
            self._export_method
        )(dest, indent=self._indent)

    def export_2d_vertical_infiltration_lines(self):
        if not self.ga.has_2d:
            logger.info(
                "[*] Model {} does not have 2D, "
                "skipping export 2d vertical infiltration lines...".format(
                    self.ga.model_name)
            )
            return
        dest = os.path.join(
            self._dest,
            constants.VERTICAL_INFILTRATION_LINES + self._extension
        )
        getattr(
            self.ga.lines.subset(
                constants.SUBSET_2D_VERTICAL_INFILTRATION
            ).reproject_to(self._epsg),
            self._export_method
        )(dest, indent=self._indent)

    def _export_groundwater_grid(self):
        if not self.ga.has_groundwater:
            logger.info(
                "[*] Model {} does not have groundwater, "
                "skipping export groundwater cells...".format(
                    self.ga.model_name
                )
            )
            return
        dest_gw = os.path.join(
            self._dest, constants.GROUNDWATER + self._extension
        )
        getattr(
            self.ga.cells.subset(
                constants.SUBSET_2D_GROUNDWATER).reproject_to(self._epsg),
            self._export_method
        )(dest_gw, indent=self._indent)

    def _export_2d_grid(self):
        if not self.ga.has_2d:
            logger.info(
                "[*] Model {} does not have 2D, "
                "skipping export 2d cells...".format(
                    self.ga.model_name)
            )
            return

        dest_ow = os.path.join(
            self._dest, constants.OPEN_WATER + self._extension
        )
        getattr(
            self.ga.cells.subset(
                constants.SUBSET_2D_OPEN_WATER).reproject_to(self._epsg),
            self._export_method
        )(dest_ow, indent=self._indent)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import numpy as np
from threedigrid.admin import constants
from threedigrid.admin.nodes.subsets import NODE_TYPE__IN_SUBSETS
from threedigrid.admin.prepare_utils import (
        db_objects_to_numpy_array_dict, add_or_update_datasets)


def as_numpy_array(array):
    if hasattr(array, 'value'):
        return array.value
    return array[()]


class PrepareNodes:

    @staticmethod
    def get_node_pks(mapping1d, id_mapper):
        # pk template
        type_code = constants.TYPE_CODE_MAP['v2_connection_nodes']

        content_pk = np.zeros(
            mapping1d['nend1d'].shape, dtype='i4'
        )

        # with argsort
        sort_idx = np.argsort(mapping1d['nend1d'])
        idx = sort_idx[np.searchsorted(
            mapping1d['nend1d'],
            id_mapper.id_mapping[
                id_mapper.obj_slices[type_code]]['seq_id'],
            sorter=sort_idx)]

        content_pk[idx] = id_mapper.id_mapping[
            id_mapper.obj_slices[type_code]]['pk']
        return content_pk

    @classmethod
    def prepare_datasource(cls, datasource, mapping1d, id_mapper, has_1d):
        if 'id' not in list(datasource.keys()):
            datasource.set(
                'id', np.arange(0, datasource['x_coordinate'].size))

        if 'content_pk' not in list(datasource.keys()) and has_1d:
            datasource.set(
                'content_pk', cls.get_node_pks(mapping1d, id_mapper))

        if 'seq_id' not in list(datasource.keys()) and has_1d:
            datasource.set('seq_id', mapping1d['nend1d'][:])

        if 'coordinates' not in list(datasource.keys()):
            datasource.set(
                'coordinates', np.array(
                    [as_numpy_array(datasource['x_coordinate']),
                     as_numpy_array(datasource['y_coordinate'])]))


class PrepareConnectionNodes:

    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        node_group = h5py_file['nodes']
        content_pk = node_group['content_pk'][:]

        connection_nodes_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.connection_nodes,
            ['pk', 'initial_waterlevel', 'storage_area'])

        add_or_update_datasets(
            node_group, connection_nodes_numpy_array_dict,
            ['initial_waterlevel', 'storage_area'],
            connection_nodes_numpy_array_dict['pk'], content_pk)


class PrepareCells:

    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        node_group = h5py_file['nodes']
        lgrmin = h5py_file['meta']['lgrmin'][()]
        nodk = h5py_file['grid_coordinate_attributes']['nodk'][()]
        nodm = h5py_file['grid_coordinate_attributes']['nodm'][()]
        nodn = h5py_file['grid_coordinate_attributes']['nodn'][()]
        ip = h5py_file['grid_coordinate_attributes']['ip'][()]
        jp = h5py_file['grid_coordinate_attributes']['jp'][()]

        node_types = h5py_file['nodes']['node_type'][:]

        pixel_width = np.zeros(nodk.shape, dtype='int')
        pixel_coords = np.full((4, nodk.shape[0]), -9999, dtype='int')
        for node_type_subset in NODE_TYPE__IN_SUBSETS['2D_ALL']:
            mask = node_types == node_type_subset
            pixel_width[mask] = lgrmin * 2 ** (nodk[mask] - 1)
            pixel_coords[0, mask] = ip[0, nodm[mask] - 1, nodk[mask] - 1] - 1
            pixel_coords[1, mask] = jp[0, nodn[mask] - 1, nodk[mask] - 1] - 1
            pixel_coords[2, mask] = ip[3, nodm[mask] - 1, nodk[mask] - 1]
            pixel_coords[3, mask] = jp[3, nodn[mask] - 1, nodk[mask] - 1]

        node_group.create_dataset('pixel_width', data=pixel_width, dtype='int')
        node_group.create_dataset(
            'pixel_coords', data=pixel_coords, dtype='int'
        )


class PrepareManholes:

    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        node_group = h5py_file['nodes']
        content_pk = node_group['content_pk'][:]

        manhole_numpy_array_dict = db_objects_to_numpy_array_dict(
            threedi_datasource.v2_manholes, [
                'pk', 'surface_level', 'display_name', 'bottom_level',
                'calculation_type', 'shape', 'drain_level', 'width',
                'manhole_indicator', 'zoom_category', 'connection_node_pk'])

        # extra field to distinguish manholes from connection nodes.
        is_manhole = np.full(len(threedi_datasource.v2_manholes), True)
        manhole_numpy_array_dict['is_manhole'] = is_manhole

        add_or_update_datasets(
            node_group, manhole_numpy_array_dict,
            ['surface_level', 'display_name', 'bottom_level',
             'calculation_type', 'shape', 'drain_level', 'width',
             'manhole_indicator', 'zoom_category', 'is_manhole'],
            manhole_numpy_array_dict['connection_node_pk'], content_pk)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
"""
Exporters
---------
At this moment there is one exporter for breach data, called
``BreachesOgrExporter``. For an overview of supported drivers call::

    >>> from threedigrid.admin.breaches.exporters import BreachesOgrExporter
    >>> from threedigrid.admin.gridadmin import GridH5Admin

    >>> f = 'gridadmin.h5'
    >>> ga = GridH5Admin(f)

    >>> # get all active breaches
    >>> active_breaches = ga.breaches.filter(kcu__eq=56)

    >>> exporter = BreachesOgrExporter(active_breaches)
    >>> exporter.supported_drivers
    >>> {u'ESRI Shapefile', u'GPKG'}

"""
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import os
import logging
from collections import OrderedDict
import six
from six.moves import range

try:
    from osgeo import ogr
except ImportError:
    ogr = None

from threedigrid.numpy_utils import reshape_flat_array
from threedigrid.geo_utils import get_spatial_reference
from threedigrid.admin.utils import KCUDescriptor
from threedigrid.admin import exporter_constants as const
from threedigrid.orm.base.exporters import BaseOgrExporter

logger = logging.getLogger(__name__)


class BreachesOgrExporter(BaseOgrExporter):
    """
    ogr exporter for breaches. See ``<instance>.supported_drivers`` to
    get a list of supported drivers
    """
    def __init__(self, breaches):
        self._breaches = breaches
        self.supported_drivers = {
            const.SHP_DRIVER_NAME,
            const.GEO_PACKAGE_DRIVER_NAME,
            const.GEOJSON_DRIVER_NAME,
        }

    def save(self, file_name, breach_data, target_epsg_code, **kwargs):
        """
        save breaches to file

        :param file_name: file name including full path
        :param breach_data: queryset
        :param target_epsg_code: desired epsg code for coords
        :param kwargs: does not take extra kwargs
        """
        assert self.driver is not None
        selection = breach_data
        kcu_dict = KCUDescriptor()
        geomtype = 0
        sr = get_spatial_reference(target_epsg_code)
        self.del_datasource(file_name)
        data_source = self.driver.CreateDataSource(file_name)
        layer = data_source.CreateLayer(
            str(os.path.basename(file_name)),
            sr,
            geomtype
        )
        fields = OrderedDict([
            ('link_id', 'int'),
            ('kcu', 'int'),
            ('kcu_descr', 'str'),
            ('cont_pk', 'int'),
        ])
        for field_name, field_type in six.iteritems(fields):
            layer.CreateField(
                ogr.FieldDefn(
                    str(field_name),
                    const.OGR_FIELD_TYPE_MAP[field_type]
                )
            )
        _definition = layer.GetLayerDefn()
        points = reshape_flat_array(selection['coordinates']).T

        for i in range(selection['id'].size):
            feature = ogr.Feature(_definition)
            point = ogr.Geometry(ogr.wkbPoint)
            point.AddPoint(points[i][0], points[i][1])
            feature.SetGeometry(point)
            feature.SetField(str("link_id"), int(selection['levl'][i]))
            kcu = selection['kcu'][i]
            feature.SetField(str("kcu"), int(kcu))
            kcu_descr = ''
            try:
                kcu_descr = kcu_dict[kcu]
            except KeyError:
                pass
            feature.SetField(str("kcu_descr"), str(kcu_descr))
            feature.SetField(str("cont_pk"), int(selection['content_pk'][i]))
            layer.CreateFeature(feature)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
"""
Main entry point for threedigrid applications.
"""
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import logging

import h5py
import numpy as np

from threedigrid.geo_utils import transform_bbox

from threedigrid.admin.lines.models import Lines
from threedigrid.admin.nodes.models import EmbeddedNodes, Nodes
from threedigrid.admin.nodes.models import Cells
from threedigrid.admin.nodes.models import Grid
from threedigrid.admin.crosssections.models import CrossSections
from threedigrid.admin.breaches.models import Breaches
from threedigrid.admin.pumps.models import Pumps
from threedigrid.admin.levees.models import Levees
from threedigrid.admin.h5py_datasource import H5pyGroup

try:
    import asyncio
    import asyncio_rpc # noqa
    asyncio_rpc_support = True
except ImportError:
    asyncio_rpc_support = False


from . import constants
import six

logger = logging.getLogger(__name__)


class GridH5Admin(object):
    """
    Parses the hdf5 gridadmin file and exposes the model instances, e.g.::

        >>> ga = GridH5Admin(file_path)
        >>> ga.nodes
        >>> ga.lines
        >>> ...
    """

    def __init__(self, h5_file_path, file_modus='r', set_props=False):
        """
        :param h5_file_path: path to the gridadmin file
        :param file_modus: mode with which to open the file
            (defaults to r=READ)
        """

        self.grid_file = h5_file_path
        self.datasource_class = H5pyGroup
        self.is_rpc = False

        if ((type(h5_file_path) in (str, bytes) and
             h5_file_path.startswith('rpc://'))):
            if not asyncio_rpc_support:
                raise Exception(
                    "Please reinstall this package with threedigrid[rpc]")

            from threedigrid.admin.rpc_datasource import H5RPCGroup, RPCFile
            self.h5py_file = RPCFile(h5_file_path, file_modus)
            self.datasource_class = H5RPCGroup
            self._grid_kwargs = {}
            self.is_rpc = True
            self.has_1d = True
        else:
            self.h5py_file = h5py.File(h5_file_path, file_modus)
            set_props = True

        if set_props:
            self._set_props()

        self._grid_kwargs = {
            'has_1d': self.has_1d
        }

    @property
    def grid(self):
        if self.is_rpc:
            raise Exception("RPC not available for grid")
        kwargs = self._grid_kwargs.copy()
        kwargs['n2dtot'] = self.get_from_meta('n2dtot')
        return Grid(
            self.datasource_class(
                self.h5py_file, 'grid_coordinate_attributes'), **kwargs
        )

    @property
    def levees(self):
        if self.is_rpc:
            raise Exception("RPC no available for levees")

        return Levees(
            self.datasource_class(
                self.h5py_file, 'levees'), **self._grid_kwargs)

    @property
    def nodes(self):
        return Nodes(
            self.datasource_class(
                self.h5py_file, 'nodes'), **self._grid_kwargs)

    @property
    def nodes_embedded(self):
        return EmbeddedNodes(
            self.datasource_class(
                self.h5py_file, 'nodes_embedded'), **self._grid_kwargs)

    @property
    def lines(self):
        return Lines(
            self.datasource_class(
                self.h5py_file, 'lines'), **self._grid_kwargs)

    @property
    def cross_sections(self):
        return CrossSections(
            self.datasource_class(
                self.h5py_file, 'cross_sections'), **self._grid_kwargs)

    @property
    def pumps(self):
        return Pumps(
            self.datasource_class(
                self.h5py_file, 'pumps'), **self._grid_kwargs)

    @property
    def breaches(self):
        return Breaches(
            self.datasource_class(
                self.h5py_file, 'breaches'), **self._grid_kwargs)

    @property
    def cells(self):
        # treated as nodes
        return Cells(
            self.datasource_class(
                self.h5py_file, 'nodes'), **self._grid_kwargs)

    @property
    def revision_hash(self):
        """mercurial revision hash of the model"""
        return self._to_str(self.h5py_file.attrs['revision_hash'])

    @property
    def revision_nr(self):
        """mercurial revision nr or id of the model"""
        return self._to_str(self.h5py_file.attrs['revision_nr'])

    @property
    def model_name(self):
        """name of the model the gridadmin file belongs to"""
        return self._to_str(self.h5py_file.attrs['model_name'])

    @property
    def epsg_code(self):
        return self._to_str(self.h5py_file.attrs['epsg_code'])

    @property
    def model_slug(self):
        return self._to_str(self.h5py_file.attrs['model_slug'])

    @property
    def threedicore_version(self):
        return self._to_str(self.h5py_file.attrs['threedicore_version'])

    @property
    def threedi_version(self):
        return self._to_str(self.h5py_file.attrs['threedi_version'])

    @property
    def has_levees(self):
        if not hasattr(self, "levees"):
            return False
        return bool(self.levees.id.size)

    def get_extent_subset(self, subset_name, target_epsg_code=''):
        """

        :param subset_name: name of the node subset for which the extent
            should be calculated.
            Valid input: '1D_all', '2D_groundwater'
        :param target_epsg_code: string representation of the desired
            output epsg code
        :return: numpy array of xy-min/xy-max pairs
        """
        if self.is_rpc:
            # Proxy function via RPC
            return self.h5py_file.get_extent_subset(
                subset_name, target_espg_code=target_epsg_code)

        attr_name = constants.SUBSET_NAME_H5_ATTR_MAP.get(subset_name.upper())
        extent = self.h5py_file.attrs.get(attr_name, None)
        if extent is None:
            raise AttributeError(
                "The grid admin file %s does not have %s attribute" % (
                    self.grid_file, attr_name)
            )
        if attr_name == constants.EXTENT_1D_KEY and not self.has_1d:
            logger.info(
                '[*] The model has no 1D returning...'
            )
            return
        if attr_name == constants.EXTENT_2D_KEY and not self.has_2d:
            logger.info(
                '[*] The model has no 2D returning...'
            )
            return

        if target_epsg_code and target_epsg_code != self.epsg_code:
            extent = transform_bbox(
                extent, self.epsg_code, target_epsg_code,
            )
        return extent

    def get_model_extent(self, target_epsg_code='', **kwargs):
        """
        get the extent of the model. Combines the extent of 1d and 2d node
        coordinates. If target_epsg_code is different from the models epsg
        code it projects the coordinates to projection of the given
        target epsg code.

        :param target_epsg_code: string representation of the desired
            output epsg code
        :param kwargs:
            extra_extent: list of xy-min/xy-max pairs that need to be
            taken into account in the model extent calculation

        :return: numpy array of xy-min/xy-max pairs

        """
        bbox = kwargs.get('extra_extent', [])
        if self.is_rpc:
            # Proxy function via RPC
            return self.h5py_file.get_model_extent(
                target_epsg_code=target_epsg_code, bbox=bbox)

        for k in constants.SUBSET_NAME_H5_ATTR_MAP.keys():
            sub_extent = self.get_extent_subset(
                subset_name=k, target_epsg_code=target_epsg_code)
            if sub_extent is None:
                continue
            bbox.append(sub_extent)

        x = np.array(bbox)[:, [0, 2]]
        y = np.array(bbox)[:, [1, 3]]

        return np.array([np.min(x), np.min(y), np.max(x), np.max(y)])

    def _set_props(self):
        if not self.is_rpc:
            for prop, value in six.iteritems(self.h5py_file.attrs):
                if prop and prop.startswith('has_'):
                    try:
                        setattr(self, prop, bool(value))
                    except AttributeError:
                        logger.warning(
                            'Can not set property {}, already exists'.format(
                                prop)
                        )
                        pass
        else:
            from threedigrid.admin.rpc_datasource import _set_property
            properties = [
                "has_1d", "has_2d", "has_breaches", "has_pumpstations"
            ]
            futures = [_set_property(self, prop) for prop in properties]
            return asyncio.gather(*futures)

    def get_from_meta(self, prop_name):
        if prop_name not in list(self.h5py_file['meta'].keys()):
            return None

        return self.h5py_file['meta'][prop_name][()]

    @staticmethod
    def _to_str(x):
        try:
            return x.decode('utf-8')
        except AttributeError:
            return x

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.h5py_file.close()

    def close(self):
        self.h5py_file.flush()
        self.h5py_file.close()



import re
from uuid import uuid4
from threedigrid.orm.base.datasource import DataSource

try:
    import asyncio_rpc # noqa
except ImportError:
    raise Exception(
        "asyncio_rpc needs to be installed when using this module,"
        "install it with: pip install asyncio_rpc")

from asyncio_rpc.models import RPCStack, RPCSubStack, RPCCall
from asyncio_rpc.client import RPCClient
from asyncio_rpc.commlayers.redis import RPCRedisCommLayer
from asyncio_rpc.serialization import msgpack


NAMESPACE = 'GRIDRESULTADMIN'
RESULT_EXPIRE_TIME = 30


async def _set_property(ga, prop):
    value = await ga.h5py_file.attrs[prop].resolve()
    setattr(ga, prop, bool(value))


class Future(object):
    def __init__(self, rpc_file, rpc_stack):
        self.rpc_file = rpc_file
        self.rpc_stack = rpc_stack

    async def resolve(self):
        """
        Resolve this future
        """
        client = await self.rpc_file.client
        return await client.rpc_call(self.rpc_stack)


class FutureResult(Future):

    async def resolve(self):
        """
        Returns by default only NetCDF results
        """
        client = await self.rpc_file.client
        return await client.rpc_call(self.rpc_stack)

    async def subscribe(
            self, only_netcdf_results=False, max_items_per_second=None):
        """
        :param rate_limit:
            maximum number of items returned per second. Cannot
            be higher than 5

        :param only_netcdf_results:
            if True, only include results that are in the NetCDF
        """
        client = await self.rpc_file.client
        if only_netcdf_results:
            self.rpc_stack.stack.append(
                RPCCall('only_netcdf_results', [], {})
            )
        if max_items_per_second:
            self.rpc_stack.stack.append(
                RPCCall('max_items_per_second', [max_items_per_second], {})
            )
        rpc_substack = RPCSubStack(**self.rpc_stack.__dict__)
        return await client.subscribe_call(rpc_substack)


class RPCAttrs:
    def __init__(self, rpc_file):
        self.rpc_file = rpc_file

    def get(self, key, *args, **kwargs):
        # Use nodes route for now
        stack = [
            RPCCall('nodes', [], {}),
            RPCCall('_datasource', [], {}),
            RPCCall('getattr', [key], {})
        ]

        rpc_stack = RPCStack(
            uuid4().hex, NAMESPACE, RESULT_EXPIRE_TIME, stack)
        return Future(self.rpc_file, rpc_stack)

    def __getitem__(self, key):
        """
        Get item from attrs
        """
        return self.get(key)


class RPCFile:
    def __init__(self, path, file_modus):
        match = re.match(r"rpc://(?P<host>\S+)/(?P<channel>\S+)", path)

        # Default value for now
        self._redis_port = 6379

        if match:
            data = match.groupdict()
            self._redis_host = data['host']
            self.pubchannel = data['channel']
        else:
            raise Exception(
                "Could not parse path %s" % path)

        self.path = path
        self.file_modus = file_modus
        self._client = None

    def filepath(self):
        return self.path

    @property
    async def client(self):
        if self._client is None:
            subchannel = uuid4().hex
            comm = await RPCRedisCommLayer.create(
                subchannel, self.pubchannel,
                host=self._redis_host,
                port=self._redis_port,
                serialization=msgpack)
            self._client = RPCClient(comm)
        return self._client

    def get_model_extent(self, target_epsg_code='', bbox=[]):
        stack = [
            RPCCall('get_model_extent', [],
                    {'target_epsg_code': target_epsg_code,
                     'bbox': bbox}),
        ]
        rpc_stack = RPCStack(
            uuid4().hex, NAMESPACE, RESULT_EXPIRE_TIME, stack)
        return Future(self, rpc_stack)

    def get_extent_subset(self, subset_name, target_espg_code=''):
        stack = [
            RPCCall('get_extent_subset', [subset_name],
                    {'target_espg_code': target_espg_code}),
        ]
        rpc_stack = RPCStack(
            uuid4().hex, NAMESPACE, RESULT_EXPIRE_TIME, stack)
        return Future(self, rpc_stack)

    def __getitem__(self, key):
        if key == 'meta':
            return self.meta
        else:
            raise Exception("Only [meta] is implemented")

    @property
    def meta(self):
        # Use nodes route for now
        stack = [
            RPCCall('nodes', [], {}),
            RPCCall('_datasource', [], {}),
            RPCCall('getattr', ['meta'], {})
        ]

        rpc_stack = RPCStack(
            uuid4().hex, NAMESPACE, RESULT_EXPIRE_TIME, stack)
        return Future(self, rpc_stack)

    @property
    def attrs(self):
        return RPCAttrs(self)


class H5RPCGroup(DataSource):
    """
    Datasource wrapper for h5py groups,
    adding meta data to all groups.
    """
    future_class = Future

    def __init__(self, rpc_file, group_name, meta=None, required=False):
        self.rpc_file = rpc_file
        self.group_name = group_name

    def get_rpc_actions(self, model, field_name=None):
        rpc_actions = []
        if model.reproject_to_epsg:
            rpc_actions.append(
                {'reproject_to': model.reproject_to_epsg}
            )
        rpc_actions += [
            x.to_dict() for x in model.slice_filters if not x._filter_as]

        if field_name is None:
            if model.only_fields:
                rpc_actions.append(
                    {'only': model.only_fields}
                )
            rpc_actions.append(
                {'data': {}}
            )
        else:
            rpc_actions.append(
                {field_name: {}})

        return rpc_actions

    def get_asyncio_rpc_stack(self, model, rpc_actions):
        model_name = model.__class__.__name__.lower()

        stack = []

        # Override for cells here
        if model_name != 'cells':
            stack.append(
                RPCCall(self.group_name, [], {})
            )

        if model_name != self.group_name:
            stack.append(
                RPCCall(model_name, [], {})
            )

        for action in rpc_actions:
            for key, value in action.items():
                if isinstance(value, list):
                    stack.append(
                        RPCCall(key, value, {})
                    )
                elif isinstance(value, dict):
                    stack.append(
                        RPCCall(key, [], value)
                    )
                else:
                    stack.append(
                        RPCCall(key, [value], {})
                    )

        return RPCStack(
            uuid4().hex, NAMESPACE, RESULT_EXPIRE_TIME, stack)

    def get_filtered_field_value(
            self, model, field_name, ts_filter=None, lookup_index=None,
            subset_index=None):
        rpc_actions = self.get_rpc_actions(
            model, field_name=field_name)
        rpc_stack = self.get_asyncio_rpc_stack(model, rpc_actions)

        return self.future_class(self.rpc_file, rpc_stack)

    def execute_query(self, model):
        rpc_actions = self.get_rpc_actions(model)
        rpc_stack = self.get_asyncio_rpc_stack(model, rpc_actions)

        return self.future_class(self.rpc_file, rpc_stack)

    def set(self, name, values):
        pass

    def getattr(self, name):
        stack = [
            RPCCall(self.group_name, [], {}),
            RPCCall('_datasource', [], {}),
            RPCCall('getattr', [name], {})
        ]

        rpc_stack = RPCStack(
            uuid4().hex, NAMESPACE, RESULT_EXPIRE_TIME, stack)
        return self.future_class(self.rpc_file, rpc_stack)

    def keys(self):
        return []

    def has_any(self, sources):
        """
        Check if the any of the sources can be found in
        the file.
        """
        # By default pretend to have all sources,
        # an exception is thrown if the source is not available
        return True


class H5RPCResultGroup(H5RPCGroup):
    future_class = FutureResult

    def __init__(self, rpc_file, group_name, netcdf_file,
                 meta=None, required=False):
        super(H5RPCResultGroup, self).__init__(
            rpc_file, group_name, meta, required)
        self.netcdf_file = netcdf_file

    def get_timestamps(self):
        stack = [
            RPCCall(self.group_name, [], {}),
            RPCCall('_datasource', [], {}),
            RPCCall('get', ['time'], {}),
            RPCCall('value', [], {})
        ]

        rpc_stack = RPCStack(
            uuid4().hex, NAMESPACE, RESULT_EXPIRE_TIME, stack)
        return self.future_class(self.rpc_file, rpc_stack)

    def get_rpc_actions(self, model, field_name=None):
        rpc_actions = super(H5RPCResultGroup, self).get_rpc_actions(
            model, field_name)

        # Check if we need to sample
        if model.timeseries_sample is not None:
            rpc_actions.insert(
                0, {'sample': model.timeseries_sample}
            )

        # Check if we need to inject timeseries_filter
        if model.timeseries_filter is not None:
            # insert action
            rpc_actions.insert(
                0, {'timeseries': model.timeseries_filter}
            )

        return rpc_actions

    def attr(self, var_name, attr_name):
        pass



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import json
import numpy as np


class NumpyEncoder(json.JSONEncoder):
    """
    convert specific numpy types to native python types
    """

    def default(self, obj):
        """
        overwrites the default method. Numpy integers and floats will be
        converted to native python types, numpy.ndarrays to python lists
        :param obj: object to encode
        :return: encoded object
        """
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, bytes):
            return obj.decode('utf-8')
        else:
            return super(NumpyEncoder, self).default(obj)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
from .constants import SHP_DRIVER_NAME, GEO_PACKAGE_DRIVER_NAME, \
    GEOJSON_DRIVER_NAME

from threedigrid.orm.base.models import Model as BaseModel
from threedigrid.orm.fields import LineArrayField
from threedigrid.orm.fields import GeomArrayField
from threedigrid.orm.filters import FILTER_MAP
from threedigrid.admin.exporter_constants import DEFAULT_EXPORT_FIELDS
from threedigrid.admin.serializers import GeoJsonSerializer


class Model(BaseModel):
    _filter_map = FILTER_MAP

    def to_centroid(self):
        """
        Returns: the filtered values with geometries
                 transformed to centroids
        """
        selection = self.data
        for field_name in self._field_names:
            field = self._get_field(field_name)
            if isinstance(field, LineArrayField) and field_name in selection:
                # Use the to_centroid function of the GeomArrayField
                # to allow different geometry to be "centroided"
                selection[field_name] =\
                    field.to_centroid(
                        selection[field_name])
        return selection

    def _is_coords(self, field_name):
        return isinstance(self._get_field(field_name), GeomArrayField)

    def _includes_coords(self, selection):
        """
        Returns: true if selection contains any data that
                 has coordinates.
        """
        found = False
        for field_name in self._field_names:
            if isinstance(self._get_field(field_name), GeomArrayField):
                found = True
                break
        return found

    def reproject_to(self, target_epsg_code):
        """
        Returns: a new class including the reprojection.

        Note: the last reproject_to is used:
            reproject_to('28992').reproject_to('4326') == reproject_to('4326')
        """
        new_class_kwargs = dict(self.class_kwargs)
        new_class_kwargs.update({
            'reproject_to_epsg': target_epsg_code})

        return self.__init_class(
            self.__class__, **new_class_kwargs)

    def _do_reproject_value(self, value, field_name, target_epsg_code):
        if target_epsg_code == self.epsg_code:
            # Already done
            return value

        field = self._get_field(field_name)

        return field.reproject(
            value, self.epsg_code, target_epsg_code)

    def __do_reproject(self, selection, target_epsg_code):
        """
        Reproject coordinate fields in selection to the
        target_epsg_code
        """
        if target_epsg_code == self.epsg_code:
            # Already done
            return selection

        # Reproject all GeomDataField's
        for field_name in self._field_names:
            field = self._get_field(field_name)
            if isinstance(field, GeomArrayField) and field_name in selection:
                # Use the reproject function of the GeomDataField
                # to allow different geometry to be reprojected
                selection[field_name] =\
                    field.reproject(
                        selection[field_name],
                        self.epsg_code, target_epsg_code)
        return selection

    def to_shape(self, file_name, **kwargs):
        self._to_ogr(SHP_DRIVER_NAME, file_name, **kwargs)

    def to_gpkg(self, file_name, **kwargs):
        self._to_ogr(
            GEO_PACKAGE_DRIVER_NAME, file_name, **kwargs)

    def to_geojson(self, file_name, **kwargs):
        if kwargs.get('use_ogr', False):
            self._to_ogr(GEOJSON_DRIVER_NAME, file_name, **kwargs)
        else:
            fields = kwargs.get(
                'fields', DEFAULT_EXPORT_FIELDS[self.__class__.__name__]
            )
            if fields == 'ALL':
                fields = self._field_names

            indent = kwargs.get('indent', None)
            serializer = GeoJsonSerializer(fields, self, indent)
            serializer.save(file_name)

    def _to_ogr(self, driver_name, file_name, **kwargs):
        exporter = self._get_exporter(driver_name)
        if not exporter:
            raise AttributeError(
                "Instance {} has no {} exporter".format(self, driver_name)
            )
        filtered = self.data
        exporter.set_driver(driver_name=driver_name)

        epsg_code = self.epsg_code
        if self.reproject_to_epsg:
            epsg_code = self.reproject_to_epsg

        exporter.save(file_name, filtered, epsg_code,
                      **kwargs)

    def _get_exporter(self, driver_name):
        for exporter in self._exporters:
            if driver_name in exporter.supported_drivers:
                return exporter
        return None



from itertools import chain


def _flatten_dict_values(d, as_set=False):
    """
    :param d: input dictionary
    :param as_set: return the flattened values as set
        (removes duplicates), default is False
    :return: list of dictionary values, if as_set as been set to True
        the set of values

    :raises TypeError if the dict values are deeper nested than [[x]]
    """

    return_func = set if as_set else list
    _values = d.values()
    if any(isinstance(i, list) for i in _values):
        return return_func(chain(*_values))
    else:
        return return_func(_values)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
"""
Models
++++++

The ``Levees`` model.
"""
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
try:
    from osgeo import ogr
except ImportError:
    ogr = None

from threedigrid.numpy_utils import reshape_flat_array
from threedigrid.geo_utils import raise_import_exception
from threedigrid.orm.models import Model
from threedigrid.orm.fields import MultiLineArrayField
from threedigrid.orm.fields import ArrayField

from threedigrid.admin.levees import exporters


class Levees(Model):

    crest_level = ArrayField()
    max_breach_depth = ArrayField()
    coords = MultiLineArrayField()

    def __init__(self, *args,  **kwargs):
        super(Levees, self).__init__(*args, **kwargs)

        self._geoms = []
        self.current_epsg = None
        self._exporters = [
            exporters.LeveeOgrExporter(self),
        ]

    @property
    def geoms(self):
        if not self._geoms:
            self.load_geoms()
        return self._geoms

    def load_geoms(self):
        """
        load levee geometries originating
        from model DB
        """

        # works on premise of ogr install
        if ogr is None:
            raise_import_exception('ogr')

        if self._geoms:
            return
        for line_coords in self.coords:
            line = ogr.Geometry(ogr.wkbLineString)
            linepoints = reshape_flat_array(line_coords).T
            for x in linepoints:
                line.AddPoint(x[0], x[1])
            self._geoms.append(line)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import numpy as np
import h5py


DT_VARIABLE = h5py.special_dtype(vlen=np.dtype('float64'))


class PrepareLevees(object):
    @staticmethod
    def prepare_datasource(h5py_file, threedi_datasource):
        v2_levee_instances = threedi_datasource.levees
        group_name = 'levees'
        if group_name not in h5py_file:
            gr = h5py_file.create_group(group_name)
        else:
            gr = h5py_file['levees']

        dset_coords = gr.create_dataset(
            'coords', (len(v2_levee_instances),), dtype=DT_VARIABLE
        )
        dset_clevel = gr.create_dataset(
            'crest_level', (len(v2_levee_instances),), dtype='f8'
        )
        dset_mbreach_depth = gr.create_dataset(
            'max_breach_depth', (len(v2_levee_instances),), dtype='f8'
        )
        dset_id = gr.create_dataset(
            'id', (len(v2_levee_instances),), dtype='i4'
        )
        for i, levee in enumerate(v2_levee_instances):
            # "F" means to flatten in column-major (Fortran- style) order
            dset_coords[i] = np.array(
                levee.the_geom.coords).flatten('F')
            dset_clevel[i] = levee.crest_level
            dset_mbreach_depth[i] = levee.max_breach_depth or -9999
            dset_id[i] = levee.pk



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
try:
    from osgeo import ogr
except ImportError:
    ogr = None

import numpy as np

from threedigrid.admin import constants
from threedigrid.geo_utils import raise_import_exception


def as_numpy_array(array):
    if hasattr(array, 'value'):
        return array.value
    return array[()]


class PrepareBreaches(object):
    """
    prepares breaches for visualization
    """
    def __init__(self):
        if ogr is None:
            raise_import_exception('ogr')

    @staticmethod
    def get_coordinates(levees, line_coords, levl):
        breaches_x = np.zeros(levl.shape, dtype='f8')
        breaches_y = np.zeros(levl.shape, dtype='f8')

        for i, line_id in enumerate(levl):
            if i == 0:
                continue
            line = ogr.Geometry(ogr.wkbLineString)

            line.AddPoint(
                line_coords[0][line_id],
                line_coords[1][line_id])
            line.AddPoint(
                line_coords[2][line_id],
                line_coords[3][line_id])

            for levee_geom in levees.geoms:
                if not levee_geom.Intersect(line):
                    continue
                intersection = levee_geom.Intersection(line)
                breaches_x[i] = intersection.GetX()
                breaches_y[i] = intersection.GetY()
                break

        return np.array([breaches_x, breaches_y])

    @classmethod
    def prepare_datasource(cls, datasource, kcu, id_mapper,
                           levees, line_coords):
        """
        :param datasource: datasource for breaches like HFD5 group
        :param kcu: array of kcu values for datasource
        :param id_mapper: threedigrid.admin.idmapper.IdMapper instance
        :param levees: threedigrid.admin.levees.model.Levees instance
        :param line_coords: coordinates from Lines instance
        :return:
        """

        # TODO: Check values below
        if 'id' not in list(datasource.keys()):
            datasource.set(
                'id', np.arange(0, datasource['levl'].size))

        if 'seq_ids' not in list(datasource.keys()):
            datasource.set(
                'seq_ids', np.arange(0, datasource['levl'].size))
        if 'content_pk' not in list(datasource.keys()):
            content_pk = np.zeros(datasource['levl'].shape, dtype='i4')
            type_code = constants.TYPE_CODE_MAP['v2_breach']
            id_mapping = id_mapper.id_mapping
            src = id_mapper.obj_slices[type_code]
            sort_by = as_numpy_array(datasource['seq_ids'])
            sort_idx = np.argsort(sort_by)
            idx = sort_idx[np.searchsorted(
                sort_by,
                id_mapping[src]['seq_id'],
                sorter=sort_idx)]

            content_pk[idx] = id_mapping[src]['pk']
            datasource.set('content_pk', content_pk)

        levl = as_numpy_array(datasource['levl'])

        if 'kcu' not in list(datasource.keys()):
            datasource.set('kcu', as_numpy_array(kcu)[levl])

        if 'coordinates' not in list(datasource.keys()) and levees:
            datasource.set(
                'coordinates',
                cls.get_coordinates(levees, as_numpy_array(line_coords), levl))



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
"""
Exporters
---------
"""
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import os
import logging
import six
from six.moves import range

try:
    from osgeo import ogr
except ImportError:
    ogr = None

try:
    from shapely import geometry as shapely_geom
except ImportError:
    shapely_geom = None


from threedigrid.geo_utils import get_spatial_reference
from threedigrid.geo_utils import raise_import_exception
from threedigrid.admin.utils import KCUDescriptor
from threedigrid.orm.base.exporters import BaseOgrExporter
from threedigrid.admin import exporter_constants as const
from threedigrid.admin.constants import LINE_BASE_FIELDS_ALL
from threedigrid.admin.constants import LINE_FIELD_NAME_MAP
from threedigrid.admin.constants import TYPE_FUNC_MAP


logger = logging.getLogger(__name__)


class LinesOgrExporter(BaseOgrExporter):
    """
    Exports to ogr formats. You need to set the driver explicitly
    before calling save()
    """
    def __init__(self, lines):
        """
        :param lines: lines.models.Lines instance
        """
        self._lines = lines
        self.supported_drivers = {
            const.GEO_PACKAGE_DRIVER_NAME,
            const.SHP_DRIVER_NAME,
            const.GEOJSON_DRIVER_NAME,
        }
        self.driver = None

    def save(self, file_name, line_data, target_epsg_code, **kwargs):
        """
        save to file format specified by the driver, e.g. shapefile

        :param file_name: name of the outputfile
        :param line_data: dict of line data
        """
        if self.driver is None:
            self.set_driver(extension=os.path.splitext(file_name)[1])

        assert self.driver is not None

        kcu_dict = KCUDescriptor()
        geomtype = 0
        sr = get_spatial_reference(target_epsg_code)

        geom_source = 'from_threedicore'
        if 'geom' in kwargs:
            geom_source = kwargs['geom']

        # shapely is needed for the LineString creation, check if installed
        if geom_source == 'from_spatialite' and shapely_geom is None:
            raise_import_exception('shapely')

        self.del_datasource(file_name)
        data_source = self.driver.CreateDataSource(file_name)
        layer = data_source.CreateLayer(
            str(os.path.basename(file_name)),
            sr,
            geomtype
        )
        fields = kwargs.get('fields', LINE_BASE_FIELDS_ALL)
        for field_name, field_type in six.iteritems(fields):
            layer.CreateField(ogr.FieldDefn(
                    str(field_name),
                    const.OGR_FIELD_TYPE_MAP[field_type])
            )
        _definition = layer.GetLayerDefn()

        node_a = line_data['line'][0]
        node_b = line_data['line'][1]
        for i in range(node_a.size):
            if geom_source == 'from_threedicore':
                line = ogr.Geometry(ogr.wkbLineString)
                line.AddPoint(line_data['line_coords'][0][i],
                              line_data['line_coords'][1][i])
                line.AddPoint(line_data['line_coords'][2][i],
                              line_data['line_coords'][3][i])
            elif geom_source == 'from_spatialite':
                linepoints = line_data['line_geometries'][i].reshape(
                    2, -1).T.astype('float64')
                line_geom = shapely_geom.LineString(linepoints)
                line = ogr.CreateGeometryFromWkt(line_geom.wkt)

            feature = ogr.Feature(_definition)
            feature.SetGeometry(line)
            for field_name, field_type in six.iteritems(fields):
                fname = LINE_FIELD_NAME_MAP.get(field_name, field_name)
                if field_name == 'kcu_descr':
                    value = ''
                    try:
                        value = str(kcu_dict[int(line_data['kcu'][i])])
                    except KeyError:
                        pass
                elif field_name == 'node_a':
                    value = TYPE_FUNC_MAP[field_type](node_a[i])
                elif field_name == 'node_b':
                    value = TYPE_FUNC_MAP[field_type](node_b[i])
                else:
                    try:
                        raw_value = line_data[fname][i]
                    except IndexError:
                        raw_value = '-9999'
                    value = TYPE_FUNC_MAP[field_type](raw_value)
                feature.SetField(str(field_name), value)

            layer.CreateFeature(feature)
            feature.Destroy()



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
from threedigrid.orm.base.timeseries_mixin import ResultMixin
from threedigrid.orm.base.timeseries_mixin import AggregateResultMixin
from threedigrid.orm.base.options import ModelMeta
import six


BASE_COMPOSITE_FIELDS = {
    's1': ['Mesh2D_s1', 'Mesh1D_s1'],
    'vol': ['Mesh2D_vol', 'Mesh1D_vol'],
    'su': ['Mesh2D_su', 'Mesh1D_su'],
    'rain': ['Mesh2D_rain', 'Mesh1D_rain'],
    'q_lat': ['Mesh2D_q_lat', 'Mesh1D_q_lat'],
    '_mesh_id': ['Mesh2DNode_id', 'Mesh1DNode_id'],
}

BASE_SUBSET_FIELDS = {
    'infiltration_rate_simple':
        {'2d_all': 'Mesh2D_infiltration_rate_simple'},
    'ucx':
        {'2d_all': 'Mesh2D_ucx'},
    'ucy':
        {'2d_all': 'Mesh2D_ucy'},
    'leak':
        {'2d_all': 'Mesh2D_leak'},
    'intercepted_volume':
        {'2d_all': 'Mesh2D_intercepted_volume'},
    'q_sss':
        {'2d_all': 'Mesh2D_q_sss'}
}


class NodesResultsMixin(ResultMixin):

    class Meta:

        # attributes for the given fields
        field_attrs = ['units', 'long_name', 'standard_name']

        # values of *_COMPOSITE_FIELDS are the variables names as known in
        # the result netCDF file. They are split into 1D and 2D subsets.
        # As threedigrid has its own subsection ecosystem they are merged
        # into a single field (e.g. the keys of *_COMPOSITE_FIELDS).

        # N.B. # fields starting with '_' are private and will not be added to
        # fields property
        composite_fields = BASE_COMPOSITE_FIELDS
        subset_fields = BASE_SUBSET_FIELDS

        lookup_fields = ('id', '_mesh_id')

    def __init__(self, **kwargs):
        """Instantiate a node with netcdf results.

        Variables stored in the netcdf and related to nodes are dynamically
        added as attributes as TimeSeriesCompositeArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super(NodesResultsMixin, self).__init__(**kwargs)


class NodesAggregateResultsMixin(AggregateResultMixin):

    class Meta(six.with_metaclass(ModelMeta)):
        base_composition = BASE_COMPOSITE_FIELDS
        base_subset_fields = BASE_SUBSET_FIELDS

        # attributes for the given fields
        field_attrs = ['units', 'long_name']

        # extra vars that will be combined with the
        # composite fields, e.g.
        # s1 --> s1_min [Mesh2D_s1_min + Mesh1D_s1_min]
        #    --> s1_max  [Mesh2D_s1_max + Mesh1D_s1_max]
        composition_vars = {
            's1': ['min', 'max', 'avg'],
            'vol': ['min', 'max', 'avg', 'sum', 'current'],
            'su': ['min', 'max', 'avg'],
            'rain': [
                'min', 'max', 'avg', 'cum', 'cum_positive', 'cum_negative'],
            'q_lat': [
                'min', 'max', 'avg', 'cum', 'cum_positive', 'cum_negative'],
            'infiltration_rate_simple': [
                'min', 'max', 'avg', 'cum', 'cum_positive', 'cum_negative'],
            'ucx': ['min', 'max', 'avg'],
            'ucy': ['min', 'max', 'avg'],
            'leak': [
                'min', 'max', 'avg', 'cum', 'cum_positive', 'cum_negative'],
            'intercepted_volume': ['min', 'max', 'avg', 'cum', 'current'],
            'q_sss': ['min', 'max', 'avg', 'cum', 'cum_positive',
                      'cum_negative']
        }

        lookup_fields = ('id', '_mesh_id')

    def __init__(self, **kwargs):
        """Instantiate a line with netcdf results.

        Variables stored in the netcdf and related to lines are dynamically
        added as attributes as TimeSeriesArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super(NodesAggregateResultsMixin, self).__init__(**kwargs)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

SHP_DRIVER_NAME = 'ESRI Shapefile'
GEO_PACKAGE_DRIVER_NAME = 'GPKG'
GEOJSON_DRIVER_NAME = 'GeoJSON'

EXPORT_SHAPEFILE = 'to_shape'
EXPORT_GEOPACKAGE = 'to_gpkg'
EXPORT_GEOJSON = 'to_geojson'

SHP_EXTENSION = '.shp'
GEO_PACKAGE_EXTENSION = '.gpkg'
GEOJSON_EXTENSION = '.json'
GEOJSON_EXTENSION2 = '.geojson'

EXTENSION_TO_DRIVER_MAP = {
    SHP_EXTENSION: SHP_DRIVER_NAME,
    GEO_PACKAGE_EXTENSION: GEO_PACKAGE_DRIVER_NAME,
    GEOJSON_EXTENSION: GEOJSON_DRIVER_NAME,
    GEOJSON_EXTENSION2: GEOJSON_DRIVER_NAME,
}

EXPORT_METHOD_TO_EXTENSION_MAP = {
    EXPORT_SHAPEFILE: SHP_EXTENSION,
    EXPORT_GEOPACKAGE: GEO_PACKAGE_EXTENSION,
    EXPORT_GEOJSON: GEOJSON_EXTENSION,
}

###############################################################################
# White list of variables and aggregation variables
AGGREGATION_OPTIONS = {
    'min',
    'max',
    'avg',
    'med',
    'cum',
    'cum_positive',
    'cum_negative'
}

###############################################################################
# result netCDF variables


# pumpstations
PUMPS_VARIABLES = ['Mesh1D_q_pump']

# breaches
BREACH_VARIABLES = ['Mesh1D_breach_depth', 'Mesh1D_breach_width']
LEVEES_VARIABLES = []



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function


class NotConsistentException(Exception):
    """Is raises when data is not consistent"""
    pass


class DriverNotSupportedError(BaseException):
    """Will be raised when trying export with an
    unknown/unsupported ogr driver"""
    pass


class OperationNotSupportedError(BaseException):
    """Will be raised when an user operation
    could result in ambiguous results """
    pass



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import json
import numpy as np

from threedigrid.admin.lines.models import Channels
from threedigrid.admin.lines.models import Culverts
from threedigrid.admin.lines.models import Pipes
from threedigrid.admin.lines.models import Orifices
from threedigrid.admin.lines.models import Weirs
from threedigrid.admin import constants
from threedigrid.geo_utils import raise_import_exception
from threedigrid.orm.base.encoder import NumpyEncoder
from six.moves import range

try:
    import geojson
except ImportError:
    geojson = None

FRICTION_CHEZY = 1
FRICTION_MANNING = 2

PIPE_FRICTION_FORMATS = {
    FRICTION_CHEZY: 'Chezy, %0.0f [m(1/2)/s]',
    FRICTION_MANNING: 'Manning, %0.3f s/[m1/3]'
}

SHAPE_RECTANGLE = 1
SHAPE_CIRCLE = 2
SHAPE_EGG = 3
SHAPE_TABULATED_RECTANGLE = 5
SHAPE_TABULATED_TRAPEZIUM = 6


def format_to_str(value, empty_value='--', format_string='%0.2f'):
    """
    Format value with format_string when evaluated as True
    else return the empty value
    """
    if not value:
        return empty_value
    return format_string % value


def friction_str(friction_value, friction_type):
    if friction_type not in PIPE_FRICTION_FORMATS:
        return 'unknown type=%d value=%f' % (friction_type, friction_value)

    return PIPE_FRICTION_FORMATS[friction_type] % friction_value


def cross_section_str(width, height, shape):
    if shape == SHAPE_TABULATED_RECTANGLE:
        return 'tabulated rect %s-%s' % (width, height)
    if shape == SHAPE_TABULATED_TRAPEZIUM:
        return 'tabulated trap %s-%s' % (width, height)

    if not width:
        width = 'NOT GIVEN'
    if not height:
        height = 'NOT GIVEN'

    lookup = {
        SHAPE_RECTANGLE: 'rectangle wxh={}mx{}m'.format(
            width, height),
        SHAPE_CIRCLE: 'circle dia={}m'.format(width),
        SHAPE_EGG: 'egg wxh={}mx{}m'.format(width, height),
    }

    return lookup[shape]


class ChannelsGeoJsonSerializer():
    def __init__(self, channels=None, data=None, indent=None):
        if geojson is None:
            raise_import_exception('geojson')
        if channels:
            assert isinstance(channels, Channels)
        self._data = data
        self._channels = channels
        self._indent = indent

    @property
    def geos(self):
        if self._channels:
            selection = self._channels.to_dict()
        else:
            selection = self._data

        geos = []
        if 'line_geometries' not in selection:
            raise ValueError(
                "Can't return data as geojson "
                "for selection without geometries")

        for i in range(selection['id'].shape[-1]):
            # Pack line_coords as [[x1, y1], [x2, y2]]
            # rounded with LONLAT_DIGITS
            lines = geojson.LineString(
                    np.round(selection['line_geometries'][i].reshape(
                        (2, -1)).T.astype('float64'),
                        constants.LONLAT_DIGITS).tolist())

            cn_meta = [
                ['object type', constants.TYPE_V2_CHANNEL],
                ['code', selection['code'][i]],
                ['calculation type',
                    constants.CALCULATION_TYPES.get(
                        selection['calculation_type'][i],
                        'unknown [%s]' % selection['calculation_type'][i]
                    )
                 ],
                ['calculation node distance',
                    selection['dist_calc_points'][i] or '--'],
                ['connection node start',
                    selection['connection_node_start_pk'][i] or '--'],
                ['connection node end',
                    selection['connection_node_end_pk'][i] or '--'],
                ['line index', selection['id'][i]],
            ]
            cn_props = dict(
                line_idx=selection['id'][i],
                id=selection['id'][i],
                object_type=constants.TYPE_V2_CHANNEL,
                props={},
                zoom_category=selection['zoom_category'][i],
                meta=cn_meta,
                )
            feat = geojson.Feature(
                geometry=lines,
                properties=cn_props,
                )
            geos.append(feat)

        return geos

    @property
    def data(self):
        geos = self.geos
        return json.dumps({
            'type': 'FeatureCollection',
            'features': geos,
        }, indent=self._indent, cls=NumpyEncoder)


class PipesGeoJsonSerializer():
    def __init__(self, pipes=None, data=None, indent=None):
        if geojson is None:
            raise_import_exception('geojson')
        if pipes:
            assert isinstance(pipes, Pipes)
        self._data = data
        self._pipes = pipes
        self._indent = indent

    @property
    def geos(self):
        if self._pipes:
            selection = self._pipes.to_dict()
        else:
            selection = self._data

        geos = []
        if 'line_coords' not in selection:
            raise ValueError(
                "Can't return data as geojson "
                "for selection without geometries")

        for i in range(selection['id'].shape[-1]):
            # Pack line_coords as [[x1, y1], [x2, y2]]
            # rounded with LONLAT_DIGITS
            lines = geojson.LineString(
                    np.round(selection['line_coords'][:, i].reshape(
                        (2, -1)), constants.LONLAT_DIGITS).tolist())

            invert_level_start_point_str = format_to_str(
                selection['invert_level_start_point'][i])
            invert_level_end_point_str = format_to_str(
                selection['invert_level_end_point'][i])

            cross_section = cross_section_str(
                selection['cross_section_width'][i],
                selection['cross_section_height'][i],
                selection['cross_section_shape'][i])

            cn_meta = [
                ['object type', constants.TYPE_V2_PIPE],
                ['display name', selection['display_name'][i]],
                ['invert level start point',
                    '%s [m MSL]' % invert_level_start_point_str],
                ['invert level end point',
                    '%s [m MSL]' % invert_level_end_point_str],
                ['friction', friction_str(
                    selection['friction_value'][i],
                    selection['friction_type'][i])],
                ['cross section', cross_section],
                ['sewerage type', constants.SEWERAGE_TYPES.get(
                    selection['sewerage_type'][i], None)],
                ['calculation type',
                    constants.CALCULATION_TYPES.get(
                        selection['calculation_type'][i],
                        'unknown [%s]' % selection['calculation_type'][i]
                    )
                 ],
                ['connection node start',
                    selection['connection_node_start_pk'][i] or '--'],
                ['connection node end',
                    selection['connection_node_end_pk'][i] or '--'],
                ['line index', selection['id'][i]],
            ]
            cn_props = dict(
                line_idx=selection['id'][i],
                id=selection['id'][i],
                object_type=constants.TYPE_V2_PIPE,
                props=dict(
                    sewerage_type=selection['sewerage_type'][i],
                ),
                zoom_category=selection['zoom_category'][i],
                meta=cn_meta,
                )
            feat = geojson.Feature(
                geometry=lines,
                properties=cn_props,
                )
            geos.append(feat)

        return geos

    @property
    def data(self):
        geos = self.geos
        return json.dumps({
            'type': 'FeatureCollection',
            'features': geos,
        }, indent=self._indent, cls=NumpyEncoder)


class WeirsGeoJsonSerializer():
    def __init__(self, weirs=None, data=None, indent=None):
        if geojson is None:
            raise_import_exception('geojson')
        if weirs:
            assert isinstance(weirs, Weirs)
        self._data = data
        self._weirs = weirs
        self._indent = indent

    @property
    def geos(self):
        if self._weirs:
            selection = self._weirs.to_dict()
            line_coords_angles = self._weirs.line_coord_angles
            line_coords_centroids = self._weirs.to_centroid()['line_coords']
        else:
            selection = self._data
            line_coords_angles = self._data['line_coord_angles']
            line_coords_centroids = self._data['line_coords_centroids']

        geos = []

        for i in range(selection['id'].shape[-1]):
            # Pack line_coords as [[x1, y1], [x2, y2]]
            # rounded with LONLAT_DIGITS
            lines = geojson.Point(
                    line_coords_centroids[:, i].tolist())

            cn_meta = [
                ['object type', constants.TYPE_V2_WEIR],
                ['display name', selection['display_name'][i]],
                ['sewerage', str(selection['sewerage'][i] == 1)],
                ['discharge coeff. pos.',
                    "%0.1f" % selection['discharge_coefficient_positive'][i]
                    or '--'],
                ['discharge coeff. neg.',
                    "%0.1f" % selection['discharge_coefficient_negative'][i]
                    or '--'],
                ['friction', friction_str(
                    selection['friction_value'][i],
                    selection['friction_type'][i])],
                ['crest type', constants.CREST_TYPES.get(
                    selection['crest_type'][i], None)],
                ['crest level',
                    '-%0.2f [m MSL]' % selection['crest_level'][i]],
                ['connection node start',
                    selection['connection_node_start_pk'][i] or '--'],
                ['connection node end',
                    selection['connection_node_end_pk'][i] or '--'],
                ['line index', selection['id'][i]],
            ]
            cn_props = dict(
                line_idx=selection['id'][i],
                id=selection['id'][i],
                angle=line_coords_angles[i],
                object_type=constants.TYPE_V2_WEIR,
                zoom_category=selection['zoom_category'][i],
                meta=cn_meta,
                )
            feat = geojson.Feature(
                geometry=lines,
                properties=cn_props,
                )
            geos.append(feat)

        return geos

    @property
    def data(self):
        geos = self.geos
        return json.dumps({
            'type': 'FeatureCollection',
            'features': geos,
        }, indent=self._indent, cls=NumpyEncoder)


class CulvertsGeoJsonSerializer():
    def __init__(self, culverts=None, data=None, indent=None):
        if geojson is None:
            raise_import_exception('geojson')
        if culverts:
            assert isinstance(culverts, Culverts)
        self._data = data
        self._culverts = culverts
        self._indent = indent

    @property
    def geos(self):
        if self._culverts:
            selection = self._culverts.to_dict()
        else:
            selection = self._data

        geos = []

        if 'line_geometries' not in selection:
            raise ValueError(
                "Can't return data as geojson "
                "for selection without geometries")

        for i in range(selection['id'].shape[-1]):
            # Pack line_coords as [[x1, y1], [x2, y2]]
            # rounded with LONLAT_DIGITS
            lines = geojson.LineString(
                    np.round(selection['line_geometries'][i].reshape(
                        (2, -1)).T.astype('float64'),
                        constants.LONLAT_DIGITS).tolist())

            invert_level_start_point_str = format_to_str(
                selection['invert_level_start_point'][i])
            invert_level_end_point_str = format_to_str(
                selection['invert_level_end_point'][i])

            cross_section = cross_section_str(
                selection['cross_section_width'][i],
                selection['cross_section_height'][i],
                selection['cross_section_shape'][i])

            cn_meta = [
                ['object type', constants.TYPE_V2_CULVERT],
                ['display name', selection['display_name'][i]],
                ['code', selection['code'][i]],
                ['calculation type',
                    constants.CALCULATION_TYPES.get(
                        selection['calculation_type'][i],
                        'unknown [%s]' % selection['calculation_type'][i]
                    )
                 ],
                ['calculation node distance',
                    selection['dist_calc_points'][i] or '--'],
                ['cross section', cross_section],
                ['friction', friction_str(
                    selection['friction_value'][i],
                    selection['friction_type'][i])],
                ['discharge coeff. pos.',
                    "%0.1f" % selection['discharge_coefficient_positive'][i]
                    or '--'],
                ['discharge coeff. neg.',
                    "%0.1f" % selection['discharge_coefficient_negative'][i]
                    or '--'],
                ['invert level start point',
                    '%s [m MSL]' % invert_level_start_point_str],
                ['invert level end point',
                    '%s [m MSL]' % invert_level_end_point_str],
                ['connection node start',
                    selection['connection_node_start_pk'][i] or '--'],
                ['connection node end',
                    selection['connection_node_end_pk'][i] or '--'],
                ['line index', selection['id'][i]],
            ]
            cn_props = dict(
                line_idx=selection['id'][i],
                id=selection['id'][i],
                object_type=constants.TYPE_V2_CULVERT,
                zoom_category=selection['zoom_category'][i],
                meta=cn_meta,
                )
            feat = geojson.Feature(
                geometry=lines,
                properties=cn_props,
                )
            geos.append(feat)

        return geos

    @property
    def data(self):
        geos = self.geos
        return json.dumps({
            'type': 'FeatureCollection',
            'features': geos,
        }, indent=self._indent, cls=NumpyEncoder)


class OrificesGeoJsonSerializer():
    def __init__(self, orifices=None, data=None, indent=None):
        if geojson is None:
            raise_import_exception('geojson')
        if orifices:
            assert isinstance(orifices, Orifices)
        self._data = data
        self._orifices = orifices
        self._indent = indent

    @property
    def geos(self):
        if self._orifices:
            selection = self._orifices.to_dict()
        else:
            selection = self._data

        geos = []

        for i in range(selection['id'].shape[-1]):
            # Pack line_coords as [[x1, y1], [x2, y2]]
            # rounded with LONLAT_DIGITS
            lines = geojson.LineString(
                    np.round(selection['line_coords'][:, i].reshape(
                        (2, -1)), constants.LONLAT_DIGITS).tolist())

            cn_meta = [
                ['object type', constants.TYPE_V2_ORIFICE],
                ['display name', selection['display_name'][i]],
                ['sewerage', str(selection['sewerage'][i] == 1)],
                ['discharge coeff. pos.',
                    "%0.1f" % selection['discharge_coefficient_positive'][i]
                    or '--'],
                ['discharge coeff. neg.',
                    "%0.1f" % selection['discharge_coefficient_negative'][i]
                    or '--'],
                ['friction', friction_str(
                    selection['friction_value'][i],
                    selection['friction_type'][i])],
                ['crest level',
                    '-%0.2f [m MSL]' % selection['crest_level'][i]],
                ['connection node start',
                    selection['connection_node_start_pk'][i] or '--'],
                ['connection node end',
                    selection['connection_node_end_pk'][i] or '--'],
                ['line index', selection['id'][i]],
            ]
            cn_props = dict(
                line_idx=selection['id'][i],
                id=None,
                object_type=constants.TYPE_V2_ORIFICE,
                zoom_category=selection['zoom_category'][i],
                meta=cn_meta,
                )
            feat = geojson.Feature(
                geometry=lines,
                properties=cn_props,
                )
            geos.append(feat)

        return geos

    @property
    def data(self):
        geos = self.geos
        return json.dumps({
            'type': 'FeatureCollection',
            'features': geos,
        }, indent=self._indent, cls=NumpyEncoder)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals
import numpy as np
import logging

try:
    import pyproj
except ImportError:
    pyproj = None

try:
    from osgeo import osr
except ImportError:
    osr = None

try:
    import mercantile
except ImportError:
    mercantile = None

try:
    import shapely
    from shapely.strtree import STRtree
    from shapely.wkt import loads
except ImportError:
    shapely = None

from threedigrid.numpy_utils import select_lines_by_bbox

logger = logging.getLogger(__name__)

MERCANTILE_EPSG_CODE = '4326'

BBOX_LEFT = 0
BBOX_TOP = 1
BBOX_RIGHT = 2
BBOX_BOTTOM = 3


def raise_import_exception(name):
    raise ImportError(
        'Could not import {}, you need to install threedigrid '
        'with the extra [geo], e.g '
        'pip install threedigrid[geo]==<version>'.format(name)
    )


def transform_xys(x_array, y_array, source_epsg, target_epsg):
    """
    Transform x_array, y_array from source_epsg_code to
    target_epsg code
    """
    if pyproj is None:
        raise_import_exception('pyproj')

    pyproj_version = pyproj.__version__.split('.')
    check_version = '2.2.0'.split('.')

    assert isinstance(x_array, np.ndarray)
    assert isinstance(y_array, np.ndarray)

    if x_array.size == 0 and y_array.size == 0:
        return np.array([[], []])
    # if threedigrid.orm.transform is None:
    #     raise ImportError('')
    projections = []
    for epsg_code in (source_epsg, target_epsg):
        if isinstance(epsg_code, bytes):
            epsg_code = epsg_code.decode('utf-8')
        epsg_str = u'epsg:{}'.format(epsg_code)

        if pyproj_version >= check_version:
            projection = pyproj.Proj(epsg_str)
        else:
            projection = pyproj.Proj(init=epsg_str)
        projections.append(projection)

    if pyproj_version >= check_version:
        reprojected = pyproj.transform(
            projections[0], projections[1], x_array, y_array, always_xy=True)
    else:
        reprojected = pyproj.transform(
            projections[0], projections[1], x_array, y_array)

    return np.array(reprojected)


def get_spatial_reference(epsg_code):
    """
    :param epsg_code: Spatial Reference System Identifier (SRID)
        as implemented by the European Petroleum Survey Group (EPSG).

    :return: an osr spatial reference instance

    :raise RuntimeError: if the import from epsg_code failed

    """
    if osr is None:
        raise_import_exception('osr')

    sr = osr.SpatialReference()
    try:
        sr.ImportFromEPSG(int(epsg_code))
        return sr
    except RuntimeError:
        logger.exception(
            "[-] Importing projection from epsg code %s failed." % epsg_code
        )
        raise


def transform_bbox(
        bbox, source_epsg_code, target_epsg_code, all_coords=False
):
    """
    Transform bbox from source_epsg_code to target_epsg_code,
    if necessary

    :returns np.array of shape 4 which represent the two coordinates:
        left, bottom and right, top.
        When `all_coords` is set to `True`, a np.array of shape 8 is given
        which represents coords of the bbox in the following order:
        left top, right top, right bottom, left bottom
    """
    if source_epsg_code != target_epsg_code:
        # XXX: Not entirely sure whether transformations between two projected
        # coordinate systems always do retain the rectangular shape of a bbox.
        # Transformations between an unprojected system (e.g. WGS84) and a
        # projected system (e.g. RDNEW) will experience distortion: the
        # resulting shape cannot be accurately represented by top left
        # and bottom right.
        source_srs = get_spatial_reference(source_epsg_code)
        target_srs = get_spatial_reference(target_epsg_code)
        if source_srs.IsProjected() != target_srs.IsProjected():
            msg = "Transforming a bbox from %s to %s is inaccurate."
            logger.warning(msg, source_epsg_code, target_epsg_code)
        # Transform to [[left, right],[top, bottom]]
        input_x = [bbox[BBOX_LEFT], bbox[BBOX_RIGHT]]
        input_y = [bbox[BBOX_TOP], bbox[BBOX_BOTTOM]]
        if all_coords:
            input_x += [bbox[BBOX_RIGHT], bbox[BBOX_LEFT]]
            input_y += [bbox[BBOX_TOP], bbox[BBOX_BOTTOM]]
        bbox_trans = np.array(
            transform_xys(
                np.array(input_x), np.array(input_y),
                source_epsg_code, target_epsg_code
            )
        )

        if all_coords:
            bbox = np.array([
                bbox_trans[0][0], bbox_trans[1][0],  # left_top
                bbox_trans[0][2], bbox_trans[1][2],  # right_top
                bbox_trans[0][1], bbox_trans[1][1],  # right_bottom
                bbox_trans[0][3], bbox_trans[1][3]  # left_bottom
            ])
        else:
            # Transform back to [left,bottom,right,top]
            bbox = np.array(
                [min(bbox_trans[0]), min(bbox_trans[1]),  # left_bottom
                 max(bbox_trans[0]), max(bbox_trans[1])  # right_top
                 ]
            )
    return bbox


def get_bbox_for_tile(tile_xyz=(0, 0, 0), target_epsg_code='4326'):
    """
    Get the bbox for a tile defined by x,y,z in epsg=target_epsg_code

    :param tile_xyz: tuple with tile x,y,z
    :param target_epsg_code: the epsg_code to reproject the bbox to.

    :return: The bbox of the tile
             (reprojected to target_epsg_code if necessary)
             Numpy array: [left, top, right, bottom]
             y-axis is inverted: bottom >= top

    Examples:
    >>> get_bbox_for_tile((10, 10, 15))
    array([-179.89013672,   85.04069252, -179.87915039,   85.04164217])

    >>> get_bbox_for_tile((10, 10, 15), '28992')
    array([  212572.58551456,  5468921.2451706 ,   212705.73721283,
            5469031.60724307])
    """
    if mercantile is None:
        raise_import_exception('mercantile')
    bbox = np.array(
        mercantile.bounds(
            mercantile.Tile(
                x=int(tile_xyz[0]), y=int(tile_xyz[1]), z=int(tile_xyz[2])))
        )

    if MERCANTILE_EPSG_CODE != target_epsg_code:
        bbox = transform_bbox(bbox, MERCANTILE_EPSG_CODE, target_epsg_code)

    return bbox


def select_points_by_bbox(points, bbox):
    """
    Return a boolean mask array for the points that are in bbox

    :param points: np.array [[x1, x2, x3...], [y1, y2, y3...]]
    :param bbox: np.array [left, top, right, bottom], bottom >= top
    :return: a boolean mask array with 'True' for the
             points that are in the bbox

    Example:
    >>> select_points_by_bbox(
    ...     points=np.array([[1, 10, 1, 10], [10, 10, 50, 50]]),
    ...     bbox=np.array([10, 10, 50, 50]))
    array([False,  True, False,  True], dtype=bool)

    >>> select_points_by_bbox(
    ...     points=np.array([[-11, 10, 10, -10], [10, 10, -11, -10]]),
    ...     bbox=np.array([-10, -10, 10, 10]))
    array([False,  True, False,  True], dtype=bool)

    >>> select_points_by_bbox(
    ...     points=np.array([[-11, 10, 10, -10], [10, 10, 0, -10]]),
    ...     bbox=np.array([-10, 0, 10, 10]))
    array([False,  True,  True, False], dtype=bool)

    >>> select_points_by_bbox(
    ...     points=np.array([[-1, 0, 0, 10], [0, 0, -11, -10]]),
    ...     bbox=np.array([0, -10, 10, 0]))
    array([False,  True, False,  True], dtype=bool)

    >>> select_points_by_bbox(
    ...     points=np.array([[100, -100], [110, -110]]),
    ...     bbox=np.array([0, -10, 10, 0]))
    array([False, False], dtype=bool)
    """
    lleft = np.array([bbox[BBOX_LEFT], bbox[BBOX_TOP]])  # lower-left
    uright = np.array([bbox[BBOX_RIGHT], bbox[BBOX_BOTTOM]])  # upper-right
    xy_points = points.T
    pre_sel_low = xy_points >= lleft
    pre_sel_up = xy_points <= uright
    return np.all((pre_sel_low & pre_sel_up), axis=1)


def select_points_by_tile(tile_xyz=(0, 0, 0), target_epsg_code='4326',
                          points=None):
    """
    Select points by a tile

    :param tile_xyz: tuple with tile x,y,z
    :param target_epsg_code: the epsg_code to reproject the bbox to
            before selecting.
    :param points: (np.array) array [[x1, x2, x3...], [y1, y2, y3...]]
    :return: A boolean mask for the points within the tile_xyz bbox.

    Example:
    >>> select_points_by_tile((10, 10, 15),
    ...     '4326',
    ...     np.array([[-179.89013670, -179.89013670],
    ...               [  85.04069252,   85.04269252]]))
    array([ True, False], dtype=bool)
    """

    assert isinstance(points, np.ndarray)

    return select_points_by_bbox(
        points, get_bbox_for_tile(tile_xyz, target_epsg_code))


def select_lines_by_tile(tile_xyz=(0, 0, 0), target_epsg_code='4326',
                         lines=None, include_intersections=False):
    """
    Select lines by a tile

    :param tile_xyz: tuple with tile x,y,z
    :param target_epsg_code: the epsg_code to reproject the bbox to
            before selecting.
    :param lines: (np.array) array:
                    x1_array=lines[0], y1_array=lines[1],
                    x2_array=lines[2], y2_array=lines[3]
           line[i] = (x1[i], y1[i]) - (x2[i], y2[i])

    :return: The bbox of the tile
             (reprojected to target_epsg_code if necessary)
    """

    assert isinstance(lines, np.ndarray)

    return select_lines_by_bbox(
        lines, get_bbox_for_tile(tile_xyz, target_epsg_code),
        include_intersections)


def select_geoms_by_geometry(geoms, geometry):
    """Build a STRtree from geoms and returns intersection with geometry

    :param geoms: list of geometries you want to search from
    :param geometry: intersection geometry
    :return: list of geoms intersecting with geometry
    """
    if shapely is None:
        raise_import_exception('shapely')

    if type(geometry) in (bytes, str):
        if isinstance(geometry, bytes):
            geometry = geometry.decode('utf-8')
        # assume wkt, try to load
        geometry = loads(geometry)

    tree = STRtree(geoms)
    # STRtree checks intersection based on bbox of the geometry only:
    # https://github.com/Toblerity/Shapely/issues/558
    intersected_geoms = tree.query(geometry)

    # reverse loop because we pop elements based on index
    for i, intersected_geom in zip(
            reversed(range(len(intersected_geoms))),
            reversed(intersected_geoms)
    ):
        if not intersected_geom.intersects(geometry):
            intersected_geoms.pop(i)
    return intersected_geoms



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

NODE_TYPE__EQ_SUBSETS = {
    '2D_OPEN_WATER': 1,
    '2D_GROUNDWATER': 2,
    '1D_NO_STORAGE': 3,
    '1D_STORAGE': 4,
    '2D_BOUNDARIES': 5,
    '2D_GROUNDWATER_BOUNDARIES': 6,
    '1D_BOUNDARIES': 7,
}

NODE_TYPE__IN_SUBSETS = dict()
NODE_TYPE__IN_SUBSETS['1D'] = [
    NODE_TYPE__EQ_SUBSETS['1D_NO_STORAGE'],
    NODE_TYPE__EQ_SUBSETS['1D_STORAGE'],
]

NODE_TYPE__IN_SUBSETS['1D_ALL'] = [
    NODE_TYPE__EQ_SUBSETS['1D_NO_STORAGE'],
    NODE_TYPE__EQ_SUBSETS['1D_STORAGE'],
    NODE_TYPE__EQ_SUBSETS['1D_BOUNDARIES']
]

NODE_TYPE__IN_SUBSETS['GROUNDWATER_ALL'] = [
        NODE_TYPE__EQ_SUBSETS['2D_GROUNDWATER'],
        NODE_TYPE__EQ_SUBSETS['2D_GROUNDWATER_BOUNDARIES']
]

NODE_TYPE__IN_SUBSETS['2D_ALL'] = [
    NODE_TYPE__EQ_SUBSETS['2D_GROUNDWATER'],
    NODE_TYPE__EQ_SUBSETS['2D_GROUNDWATER_BOUNDARIES'],
    NODE_TYPE__EQ_SUBSETS['2D_BOUNDARIES'],
    NODE_TYPE__EQ_SUBSETS['2D_OPEN_WATER']
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from __future__ import absolute_import
from setuptools import setup, find_packages

import sys
import codecs
import re
import os

python_27 = sys.version_info <= (2, 7)

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

lightweight_requirements = [
    'numpy>=1.13',
    'h5py>=2.7.1',
    'six',
]

results_requirements = [
    'cftime>=1.0.1'
]

# for extra 'geo'
geo_requirements = [
    'Click>=6.0',
    'mercantile>=1.0.1',
    'pyproj>=1.9.5.1',
    'Shapely>=1.6.4',
    'geojson>=2.3.0',
]

rpc_requirements = [
    'asyncio-rpc>=0.1.10',
]

setup_requirements = []

test_requirements = ['pytest==3.4.1']

setup(
    author="Lars Claussen",
    author_email='lars.claussen@nelen-schuurmans.nl',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        "Programming Language :: Python :: 3",
        'Topic :: Scientific/Engineering',
    ],
    description="Python package for the threedigrid administration",
    entry_points={
        'console_scripts': [
            '3digrid_explore=threedigrid.management.commands.kick:kick_start',
            '3digrid_export=threedigrid.management.commands.kick:export_to',
        ],
    },
    install_requires=lightweight_requirements,
    license="BSD license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='threedigrid',
    name='threedigrid',
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    extras_require={
        'geo': geo_requirements,
        'results': results_requirements,
        'rpc': rpc_requirements
    },
    url='https://github.com/nens/threedigrid',
    version=find_version("threedigrid", "__init__.py"),
    zip_safe=False,
)



from __future__ import absolute_import
from abc import ABCMeta
from abc import abstractmethod
import os
import logging
import six

try:
    from osgeo import ogr
except ImportError:
    ogr = None

try:
    from osgeo import gdal
except ImportError:
    gdal = None

from threedigrid.orm import constants
from threedigrid.orm.base.exceptions import DriverNotSupportedError
from threedigrid.geo_utils import raise_import_exception

logger = logging.getLogger(__name__)


class BaseExporterObject(six.with_metaclass(ABCMeta)):

    @abstractmethod
    def save(self):
        """
        exporter objects must implement the save method
        """
        pass


class BaseOgrExporter(BaseExporterObject):

    driver = None

    def __init__(self):
        if ogr is None:
            raise_import_exception('ogr')
        if gdal is None:
            raise_import_exception('gdal')

    def set_driver(self, driver_name='', extension=''):
        assert any((driver_name, extension)), \
            'either driver_name or extension must be given'
        if not driver_name:
            driver_name = constants.EXTENSION_TO_DRIVER_MAP[extension.lower()]
        if all([driver_name == constants.GEO_PACKAGE_DRIVER_NAME,
                int(gdal.VersionInfo('VERSION_NUM')) < 2000000]):
            logger.error('Requires GDAL >= 2.0(dev)')
            raise DriverNotSupportedError('Requires GDAL >= 2.0(dev)')
        self.driver = ogr.GetDriverByName(str(driver_name))

    @property
    def driver_name(self):
        if self.driver:
            return self.driver.GetName()

    def del_datasource(self, file_name):
        if not os.path.exists(file_name):
            return
        logger.info('[*] Replacing %s ...', file_name)
        self.driver.DeleteDataSource(str(file_name))



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function

from __future__ import absolute_import
import numpy as np
import json
from threedigrid.admin.breaches.models import Breaches
from threedigrid.admin import constants
from threedigrid.geo_utils import raise_import_exception
from threedigrid.orm.base.encoder import NumpyEncoder
from six.moves import range

try:
    import geojson
except ImportError:
    geojson = None


class BreachesGeoJsonSerializer():
    def __init__(self, breaches=None, data=None, indent=None):
        if geojson is None:
            raise_import_exception('geojson')
        if breaches:
            assert isinstance(breaches, Breaches)
        self._data = data
        self._breaches = breaches
        self._indent = indent

    @property
    def geos(self):
        if self._breaches:
            selection = self._breaches.to_dict()
        else:
            selection = self._data

        geos = []
        if 'coordinates' not in selection:
            raise ValueError(
                "Can't return data as geojson "
                "for selection without geometries")
        for i in range(selection['id'].shape[-1]):
            pt = geojson.Point(
                    np.round(
                        selection['coordinates'][:, i],
                        constants.LONLAT_DIGITS).tolist())
            breach_meta = [
                    ['object_type', constants.TYPE_V2_BREACH],
                    ['line_idx', int(selection['levl'][i])],
                    ['breach_idx', int(selection['id'][i])],
                    ]
            breach_props = dict(
                object_type=constants.TYPE_V2_BREACH,
                props={},
                line_idx=int(selection['levl'][i]),
                breach_idx=int(selection['id'][i]),
                meta=breach_meta,
                )
            feat = geojson.Feature(
                geometry=pt,
                properties=breach_props,
                )
            geos.append(feat)

        return geos

    @property
    def data(self):
        return json.dumps({
            'type': 'FeatureCollection',
            'features': self.geos,
        }, indent=self._indent, cls=NumpyEncoder)
